import os

import matplotlib.pyplot as plt
import numpy as np
import strax
import straxen


@straxen.mini_analysis(requires=("raw_records",), warn_beyond_sec=5)
def plot_pulses_tpc(
    context,
    raw_records,
    run_id,
    time_range=None,
    plot_hits=False,
    plot_median=False,
    max_plots=20,
    store_pdf=False,
    path="",
):
    """Mini-analyis to plot pulses for the specified list of records. You have to provide a a run-id
    for which pulses should be plotted. You can use the same arguments as for get_array to select a
    specific time range or data (see also further ).

    In addition you can provide the following arguments:

    :param plot_hits: If True plot hit boundaries including the left
        and right extension as orange shaded regions.
    :param plot_median: If true plots pulses sample median as dotted
        line.
    :param max_plots: Limits the number of figures. If you would like
        to plot more pulses you should put the plots in a PDF.
    :param store_pdf: If true figures are put to a PDF instead of
        plotting them to your notebook. The file name is automatically
        generated including the time range and run_id.
    :param path: Relative path where the PDF should be stored. By default
        it is the directory of the notebook.

    """
    plot_pulses(
        context, raw_records, run_id, time_range, plot_hits, plot_median, max_plots, store_pdf, path
    )


@straxen.mini_analysis(requires=("raw_records_mv",), warn_beyond_sec=5)
def plot_pulses_mv(
    context,
    raw_records_mv,
    run_id,
    time_range=None,
    plot_hits=False,
    plot_median=False,
    max_plots=20,
    store_pdf=False,
    path="",
):
    """Mini-analyis to plot pulses for the specified list of records. You have to provide a a run-id
    for which pulses should be plotted. You can use the same arguments as for get_array to select a
    specific time range or data (see also further ).

    In addition you can provide the following arguments:

    :param plot_hits: If True plot hit boundaries including the left
        and right extension as orange shaded regions.
    :param plot_median: If true plots pulses sample median as dotted
        line.
    :param max_plots: Limits the number of figures. If you would like
        to plot more pulses you should put the plots in a PDF.
    :param store_pdf: If true figures are put to a PDF instead of
        plotting them to your notebook. The file name is automatically
        generated including the time range and run_id.
    :param path: Relative path where the PDF should be stored. By default
        it is the directory of the notebook.

    """
    plot_pulses(
        context,
        raw_records_mv,
        run_id,
        time_range,
        plot_hits,
        plot_median,
        max_plots,
        store_pdf,
        path,
        detector_ending="_mv",
    )


@straxen.mini_analysis(requires=("raw_records_nv",), warn_beyond_sec=5)
def plot_pulses_nv(
    context,
    raw_records_nv,
    run_id,
    time_range=None,
    plot_hits=False,
    plot_median=False,
    max_plots=20,
    store_pdf=False,
    path="",
):
    """Mini-analyis to plot pulses for the specified list of records. You have to provide a a run-id
    for which pulses should be plotted. You can use the same arguments as for get_array to select a
    specific time range or data (see also further ).

    In addition you can provide the following arguments:

    :param plot_hits: If True plot hit boundaries including the left
        and right extension as orange shaded regions.
    :param plot_median: If true plots pulses sample median as dotted
        line.
    :param max_plots: Limits the number of figures. If you would like
        to plot more pulses you should put the plots in a PDF.
    :param store_pdf: If true figures are put to a PDF instead of
        plotting them to your notebook. The file name is automatically
        generated including the time range and run_id.
    :param path: Relative path where the PDF should be stored. By default
        it is the directory of the notebook.

    """
    plot_pulses(
        context,
        raw_records_nv,
        run_id,
        time_range,
        plot_hits,
        plot_median,
        max_plots,
        store_pdf,
        path,
        detector_ending="_nv",
    )


def plot_pulses(
    context,
    raw_records,
    run_id,
    time_range,
    plot_hits=False,
    plot_median=False,
    max_plots=20,
    store_pdf=False,
    path="",
    detector_ending="",
):
    """Plots nveto pulses for a list of records.

    :param context: Context to be used.
    :param plot_hits: If True plot hit boundaries including the left and right extension as orange
        shaded regions.
    :param plot_median: If true plots pulses sample median as dotted line.
    :param max_plots: Limits the number of figures. If you would like to plot more pulses you should
        put the plots in a PDF.
    :param store_pdf: If true figures are put to a PDF instead of plotting them to your notebook.
        The file name is automatically generated including the time range and run_id.
    :param path: Relative path where the PDF should be stored. By default it is the directory of the
        notebook.
    :param detector_ending: Ending of the corresponding detector. Empty string for TPC '_nv' for
        neutron-veto and '_mv' muon-veto.

    """
    # Register records plugin to get settings
    p = context.get_single_plugin(run_id, "records" + detector_ending)

    # Compute strax baseline and baseline_rms:
    records = strax.raw_to_records(raw_records)
    records = strax.sort_by_time(records)
    strax.zero_out_of_bounds(records)

    baseline_key = [key for key in p.config.keys() if "baseline_samples" in key][0]
    baseline_samples = getattr(p, baseline_key)

    strax.baseline(records, baseline_samples=baseline_samples, flip=True)

    nfigs = 1
    if store_pdf and time_range is None:
        raise ValueError(f"Specify time range!")
    if store_pdf:
        from matplotlib.backends.backend_pdf import PdfPages

        fname = f"pulses_{run_id}_{time_range[0]}_{time_range[1]}.pdf"
        fname = os.path.join(path, fname)
        pdf = PdfPages(fname)

    hits = None  # needed for delete if false

    for inds in _yield_pulse_indices(raw_records):
        # Grouped our pulse so now plot:
        rr_pulse = raw_records[inds]
        r_pulse = records[inds]

        fig, axes = straxen.plot_single_pulse(rr_pulse, run_id)
        if detector_ending == "_nv":
            # We only store for the nv digitizer baseline values:
            axes.axhline(
                rr_pulse[0]["baseline"],
                ls="dashed",
                color="k",
                label=f'D. Bas.: {rr_pulse[0]["baseline"]} ADC',
            )

        baseline = r_pulse[0]["baseline"]
        baseline_rms = r_pulse[0]["baseline_rms"]
        axes.axhline(
            baseline,
            ls="solid",
            color="k",
            label=f"Strax Bas. +/-RMS:\n ({baseline:.2f}+/-{baseline_rms:.2f}) ADC",
        )
        xlim = axes.get_xlim()
        axes.fill_between(
            xlim,
            [baseline + baseline_rms] * 2,
            [baseline - baseline_rms] * 2,
            color="gray",
            alpha=0.4,
        )

        # check type of p.hit_thresholds
        if isinstance(p.hit_thresholds, int):
            thr = p.hit_thresholds
        elif isinstance(p.hit_thresholds, np.ndarray):
            thr = p.hit_thresholds[rr_pulse["channel"]][0]

        if plot_median:
            # Plot median if asked.
            # Have to make pulse again:
            pulse = straxen.matplotlib_utils._make_pulse(rr_pulse)
            median = np.median(pulse)
            axes.axhline(median, ls="dotted", color="k", label=f"Median Bas.: {median:.0f} ADC")

            axes.axhline(median - thr, ls="dotted", color="orange")

        if plot_hits:
            min_amplitude = thr

            axes.axhline(baseline - min_amplitude, color="orange", label="Hitfinder threshold")

            hits = strax.find_hits(r_pulse, min_amplitude=min_amplitude)
            if detector_ending != "_he":
                # We don't have 'save_outside_hits_he' at all!
                le, re = getattr(p, "save_outside_hits" + detector_ending)
            else:
                le, re = getattr(p, "save_outside_hits")
            start = (hits["time"] - r_pulse[0]["time"]) / r_pulse[0]["dt"] - le
            end = (strax.endtime(hits) - r_pulse[0]["time"]) / r_pulse[0]["dt"] + re

            ylim = axes.get_ylim()
            for s, e in zip(start, end):
                plt.fill_between((s, e), *ylim, alpha=0.2, color="orange")
            axes.set_ylim(*ylim)

        plt.legend()
        axes.set_xlim(*xlim)

        if store_pdf:
            plt.close()
            pdf.savefig(fig)

        nfigs += 1
        if max_plots is not None and nfigs > max_plots:
            break

    if store_pdf:
        pdf.close()
    del records, hits


def _yield_pulse_indices(records):
    """Function which yields indices of records which are within a pulse.

    Note:
        Only finds fragments of the pulse if record_i == 0 is within list
        of records.
    :yields: indices of fragments to make the corresponding pulse.

    """
    # Get record links and find start indicies:
    _, next_ri = strax.record_links(records)
    start_ri = np.where(records["record_i"] == 0)[0]

    # Loop over pulse start_ri, group fragments by pulses yield for plot:
    for ri in start_ri:
        # Buffer for indices:
        inds = []

        tries = 1
        max_tries = 5000
        while ri != -1:
            inds.append(ri)
            ri = next_ri[ri]

            tries += 1
            if tries > max_tries:
                raise ValueError(
                    "Tried more than 5000 times to find subsequent record. Am I stuck in a loop?"
                )
        yield inds



import warnings

import bokeh
import bokeh.plotting as bklt
import numba
import numpy as np
import strax
import straxen
from straxen.analyses.holoviews_waveform_display import (
    _hvdisp_plot_records_2d,
    hook,
    plot_record_polygons,
    get_records_matrix_in_window,
)  # noqa

# Default legend, unknow, S1 and S2
LEGENDS = ("Unknown", "S1", "S2")
straxen._BOKEH_CONFIGURED_NOTEBOOK = False


@straxen.mini_analysis(
    requires=("event_basics", "peaks", "peak_basics", "peak_positions"), warn_beyond_sec=0.05
)
def event_display_interactive(
    events,
    peaks,
    to_pe,
    run_id,
    context,
    bottom_pmt_array=True,
    only_main_peaks=False,
    only_peak_detail_in_wf=False,
    plot_all_pmts=False,
    plot_record_matrix=False,
    plot_records_threshold=10,
    colors=("gray", "blue", "green"),
    yscale=("linear", "linear", "linear"),
    log=True,
):
    """Interactive event display for XENONnT. Plots detailed main/alt S1/S2, bottom and top PMT hit
    pattern as well as all other peaks in a given event.

    :param bottom_pmt_array: If true plots bottom PMT array hit-pattern.
    :param only_main_peaks: If true plots only main peaks into detail
        plots as well as PMT arrays.
    :param only_peak_detail_in_wf: Only plots main/alt S1/S2 into
        waveform. Only plot main peaks if only_main_peaks is true.
    :param plot_all_pmts: Bool if True, colors switched off PMTs instead
        of showing them in gray, useful for graphs shown in talks.
    :param plot_record_matrix: If true record matrix is plotted below.
        waveform.
    :param plot_records_threshold: Threshold at which zoom level to display
        record matrix as polygons. Larger values may lead to longer
        render times since more polygons are shown.
    :param colors: Colors to be used for peaks. Order is as peak types,
        0 = Unknown, 1 = S1, 2 = S2. Can be any colors accepted by bokeh.
    :param yscale: Defines scale for main/alt S1 == 0, main/alt S2 == 1,
        waveform plot == 2. Please note, that the log scale can lead to funny
        glyph renders for small values.
    :param log: If true color scale is used for hitpattern plots.

    example::

        from IPython.core.display import display, HTML
        display(HTML("<style>.container { width:80% !important; }</style>"))
        import bokeh.plotting as bklt
        fig = st.event_display_interactive(run_id, time_range=(event['time'], event['endtime']))
        bklt.show(fig)

    :raises:
        Raises an error if the user queries a time range which contains
        more than a single event.

    :return: bokeh.plotting.figure instance.

    """
    st = context

    if len(yscale) != 3:
        raise ValueError(f'"yscale" needs three entries, but you passed {len(yscale)}.')

    if not hasattr(st, "_BOKEH_CONFIGURED_NOTEBOOK"):
        st._BOKEH_CONFIGURED_NOTEBOOK = True
        # Configure show to show notebook:
        from bokeh.io import output_notebook

        output_notebook()

    if len(events) != 1:
        raise ValueError(
            "The time range you specified contains more or"
            " less than a single event. The event display "
            " only works with individual events for now."
        )

    if peaks.shape[0] == 0:
        raise ValueError("Found an event without peaks this should not had have happened.")

    # Select main/alt S1/S2s based on time and endtime in event:
    m_other_peaks = np.ones(len(peaks), dtype=bool)  # To select non-event peaks
    endtime = strax.endtime(peaks)

    signal = {}
    if only_main_peaks:
        s1_keys = ["s1"]
        s2_keys = ["s2"]
        labels = {"s1": "S1", "s2": "S2"}
    else:
        s1_keys = ["s1", "alt_s1"]
        s2_keys = ["s2", "alt_s2"]
        labels = {"s1": "MS1", "alt_s1": "AS1", "s2": "MS2", "alt_s2": "AS2"}

    for s_x in labels.keys():
        # Loop over Main/Alt Sx and get store S1/S2 Main/Alt in signals,
        # store information about other peaks as "m_other_peaks"
        m = (peaks["time"] == events[f"{s_x}_time"]) & (endtime == events[f"{s_x}_endtime"])
        signal[s_x] = peaks[m]
        m_other_peaks &= ~m

    # Detail plots for main/alt S1/S2:
    fig_s1, fig_s2 = plot_detail_plot_s1_s2(
        signal,
        s1_keys,
        s2_keys,
        labels,
        colors,
        yscale=yscale[:2],
    )

    # PMT arrays:
    if not only_main_peaks:
        # Plot all keys into both arrays:
        top_array_keys = s2_keys + s1_keys
        # Plot S1 first then S2 in the bottom array:
        bottom_array_keys = s1_keys + s2_keys
    else:
        top_array_keys = s2_keys
        bottom_array_keys = s1_keys

    fig_top, fig_bottom = plot_pmt_arrays_and_positions(
        top_array_keys,
        bottom_array_keys,
        signal,
        to_pe,
        labels,
        plot_all_pmts,
        log=log,
    )

    m_other_s2 = m_other_peaks & (peaks["type"] == 2)
    if np.any(m_other_s2) and not only_main_peaks:
        # Now we have to add the positions of all the other S2 to the top pmt array
        # if not only main peaks.
        fig_top, plots = plot_posS2s(
            peaks[m_other_s2], label="OS2s", fig=fig_top, s2_type_style_id=2
        )
        for plot in plots:
            plot.visible = False

    # Main waveform plot:
    if only_peak_detail_in_wf:
        # If specified by the user only plot main/alt S1/S2
        peaks = peaks[~m_other_peaks]

    waveform = plot_event(peaks, signal, labels, events[0], colors, yscale=yscale[-1])

    # Create tile:
    title = _make_event_title(events[0], run_id)

    # Put everything together:
    if bottom_pmt_array:
        upper_row = [fig_s1, fig_s2, fig_top, fig_bottom]
    else:
        upper_row = [fig_s1, fig_s2, fig_top]

    upper_row = bokeh.layouts.Row(
        children=upper_row,
        sizing_mode="scale_width",
    )

    plots = bokeh.layouts.gridplot(
        children=[upper_row, waveform],
        sizing_mode="scale_width",
        ncols=1,
        merge_tools=True,
        toolbar_location="above",
    )
    event_display = bokeh.layouts.Column(
        children=[title, plots],
        sizing_mode="scale_width",
        max_width=1600,
    )

    # Add record matrix if asked:
    if plot_record_matrix:
        if st.is_stored(run_id, "records"):
            # Check if records can be found and load:
            r = st.get_array(
                run_id, "records", time_range=(events[0]["time"], events[0]["endtime"])
            )
        elif st.is_stored(run_id, "raw_records"):
            warnings.warn(
                f"Cannot find records for {run_id}, making them from raw_records instead."
            )
            p = st.get_single_plugin(run_id, "records")
            r = st.get_array(
                run_id, "raw_records", time_range=(events[0]["time"], events[0]["endtime"])
            )
            r = p.compute(r, events[0]["time"], events[0]["endtime"])["records"]
        else:
            warnings.warn(
                f"Can neither find records nor raw_records for run {run_id}, proceed without record"
                " matrix."
            )
            plot_record_matrix = False

    if plot_record_matrix:
        straxen._BOKEH_X_RANGE = None
        # First get hook to for x_range:
        x_range_hook = lambda plot, element: hook(plot, x_range=straxen._BOKEH_X_RANGE, debug=False)

        # Create datashader plot:
        wf, record_points, time_stream = _hvdisp_plot_records_2d(
            records=r,
            to_pe=to_pe,
            t_reference=peaks[0]["time"],
            event_range=(waveform.x_range.start, waveform.x_range.end),
            config=st.config,
            hooks=[x_range_hook],
            tools=[],
        )
        # Create record polygons:
        polys = plot_record_polygons(record_points)
        records_in_window = polys.apply(
            get_records_matrix_in_window, streams=[time_stream], time_slice=plot_records_threshold
        )

        # Render plot to initialize x_range:
        import holoviews as hv
        import panel

        _ = hv.render(wf)
        # Set x-range of event plot:
        bokeh_set_x_range(waveform, straxen._BOKEH_X_RANGE, debug=False)
        event_display = panel.Column(
            event_display,
            wf * records_in_window,
            sizing_mode="scale_width",
        )

    return event_display


def plot_detail_plot_s1_s2(
    signal,
    s1_keys,
    s2_keys,
    labels,
    colors,
    title=["Main/Alt S1", "Main/Alt S2"],
    yscale=("linear", "linear"),
):
    """Function to plot the main/alt S1/S2 peak details.

    :param signal: Dictionary containing the peak information.
    :param s1_keys: S1 keys to be plotted e.g. with and without alt S1
    :param s2_keys: Same but for S2
    :param labels: Labels to be used for Peaks
    :param colors: Colors to be used
    :param yscale: Tuple with axis scale type.
    :return: S1 and S2 bokeh figure.

    """
    # First we create figure then we loop over figures and plots and
    # add drawings:
    fig_s1 = straxen.bokeh_utils.default_fig(
        title=title[0],
        y_axis_type=yscale[0],
    )
    fig_s2 = straxen.bokeh_utils.default_fig(
        title=title[1],
        y_axis_type=yscale[1],
    )

    for fig, peak_types in zip([fig_s1, fig_s2], (s1_keys, s2_keys)):
        # Loop over fig and corresponding peak keys
        for peak_type in peak_types:
            if "s2" in peak_type:
                # If S2 use µs as units
                time_scalar = 1000  # ns
                unit = "µs"
            else:
                time_scalar = 1  # ns
                unit = "ns"
            if signal[peak_type].shape[0]:
                # If signal exists, plot:
                fig, plot = plot_peak_detail(
                    signal[peak_type],
                    time_scalar=time_scalar,
                    label=labels[peak_type],
                    unit=unit,
                    fig=fig,
                    colors=colors,
                )
                if "alt" in peak_type:
                    # Not main S1/S2, so make peak invisible
                    plot.visible = False
    return fig_s1, fig_s2


def plot_pmt_arrays_and_positions(
    top_array_keys, bottom_array_keys, signal, to_pe, labels, plot_all_pmts, log=True
):
    """Function which plots the Top and Bottom PMT array.

    :return: fig_top, fig_bottom

    """
    # Same logic as for detailed Peaks, first make figures
    # then loop over figures and data and populate figures with plots
    fig_top = straxen.bokeh_utils.default_fig(title="Top array")
    fig_bottom = straxen.bokeh_utils.default_fig(title="Bottom array")

    for pmt_array_type, fig, peak_types in zip(
        ["top", "bottom"], [fig_top, fig_bottom], [top_array_keys, bottom_array_keys]
    ):
        for ind, k in enumerate(peak_types):
            # Loop over peaks enumerate them since we plot all Peaks
            # Main/ALt S1/S2 into the PMT array, but only the first one
            # Should be visible.
            if not signal[k].shape[0]:
                # alt S1/S2 does not exist so go to next.
                continue

            fig, plot, _ = plot_pmt_array(
                signal[k][0],
                pmt_array_type,
                to_pe,
                plot_all_pmts=plot_all_pmts,
                label=labels[k],
                fig=fig,
                log=log,
            )
            if ind:
                # Not main S1 or S2
                plot.visible = False

            if pmt_array_type == "top" and "s2" in k:
                # In case of the top PMT array we also have to plot the S2 positions:
                fig, plots = plot_posS2s(
                    signal[k][0], label=labels[k], fig=fig, s2_type_style_id=ind
                )
                if ind:
                    # Not main S2
                    for plot in plots:
                        plot.visible = False

    return fig_top, fig_bottom


def plot_event(peaks, signal, labels, event, colors, yscale="linear"):
    """Wrapper for plot peaks to highlight main/alt. S1/S2.

    :param peaks: Peaks in event
    :param signal: Dictionary containing main/alt. S1/S2
    :param labels: dict with labels to be used
    :param event: Event to set correctly x-ranges.
    :param colors: Colors to be used for unknown, s1 and s2 signals.
    :param yscale: string of yscale type.
    :return: bokeh.plotting.figure instance

    """
    waveform = plot_peaks(peaks, time_scalar=1000, colors=colors, yscale=yscale)
    # Highlight main and alternate S1/S2:
    start = peaks[0]["time"]
    end = strax.endtime(peaks)[-1]
    # Workaround did not manage to scale via pixels...
    ymax = np.max((peaks["data"].T / peaks["dt"]).T)
    ymax -= 0.1 * ymax
    for s, p in signal.items():
        if p.shape[0]:
            pos = (p[0]["center_time"] - start) / 1000
            main = bokeh.models.Span(
                location=pos,
                dimension="height",
                line_alpha=0.6,
            )
            vline_label = bokeh.models.Label(
                x=pos,
                y=ymax,
                angle=np.pi / 2,
                text=labels[s],
            )
            if "alt" in s:
                main.line_dash = "dotted"
            else:
                main.line_dash = "dashed"
            waveform.add_layout(main)
            waveform.add_layout(vline_label)

    # Get some meaningful x-range limit to 10% left and right extending
    # beyond first last peak, clip at event boundary.
    length = (end - start) / 10**3

    waveform.x_range.start = max(-0.1 * length, (event["time"] - start) / 10**3)
    waveform.x_range.end = min(1.1 * length, (event["endtime"] - start) / 10**3)
    return waveform


@straxen.mini_analysis(
    requires=("peaks", "peak_basics", "peak_positions"),
    default_time_selection="touching",
    warn_beyond_sec=10,
)
def peaks_display_interactive(
    peaks,
    to_pe,
    run_id,
    context,
    times=[],
    center_times=[],
    bottom_pmt_array=True,
    plot_all_pmts=False,
    colors=("gray", "blue", "green"),
    yscale=("linear", "linear", "linear"),
    log=True,
    _provide_peaks=False,
):
    """Interactive events display for XENONnT. Plots detailed waveform, bottom and top PMT hit
    pattern for selected center time.

    :param bottom_pmt_array: If true plots bottom PMT array hit-pattern.
    :param colors: Colors to be used for peaks. Order is as peak types, 0 = Unknown, 1 = S1, 2 = S2.
        Can be any colors accepted by bokeh.
    :param yscale: Defines scale for main/alt S1 == 0, main/alt S2 == 1, waveform plot == 2. Please
        note, that the log scale can lead to funny glyph renders for small values.
    :param log: If true color scale is used for hitpattern plots.
    :return: bokeh.plotting.figure instance.

    """
    st = context

    if len(yscale) != 3:
        raise ValueError(f'"yscale" needs three entries, but you passed {len(yscale)}.')

    if not hasattr(st, "_BOKEH_CONFIGURED_NOTEBOOK"):
        st._BOKEH_CONFIGURED_NOTEBOOK = True
        # Configure show to show notebook:
        from bokeh.io import output_notebook

        output_notebook()

    if len(times) and len(center_times):
        raise ValueError("Please specify either times or center_times, not both.")
    if len(times):
        unique = np.unique(times)
        field = "time"
    elif len(center_times):
        unique = np.unique(center_times)
        field = "center_time"
    else:
        unique = []
        warnings.warn("No times or center_times specified, will not plot any peak in detail.")

    signal = {}
    s1_keys = []
    s2_keys = []
    labels = {}
    found_s1 = False
    found_s2 = False
    for ind, t in enumerate(unique):
        _p = peaks[peaks[field] == t]
        if not _p.shape[0]:
            raise ValueError(f"Could not find peak at center time {t}.")
        if len(_p) > 1:
            warnings.warn(f"Found multiple peaks at {field} {t}, using the first one.")
        p = _p[0]
        if np.isin(p["type"], [0, 1, 3]):
            key = f"s{p['type']}_{ind}"
            if found_s1:
                key = f"alt_{key}"
            found_s1 = True
            s1_keys.append(key)
        else:
            key = f"s{p['type']}_{ind}"
            if found_s2:
                key = f"alt_{key}"
            assert "s2" in key, "Only S2 peaks are allowed here."
            found_s2 = True
            s2_keys.append(key)
        signal[key] = _p
        labels[key] = f"{ind}"

    # Detail plots for selected peaks
    fig_s1, fig_s2 = plot_detail_plot_s1_s2(
        signal,
        s1_keys,
        s2_keys,
        labels,
        colors,
        title=["S0, S1 and others", "S2 and others"],
        yscale=yscale[:2],
    )

    # PMT arrays:
    # Plot all keys into both arrays:
    top_array_keys = s2_keys + s1_keys
    # Plot S1 first then S2 in the bottom array:
    bottom_array_keys = s1_keys + s2_keys

    fig_top, fig_bottom = plot_pmt_arrays_and_positions(
        top_array_keys,
        bottom_array_keys,
        signal,
        to_pe,
        labels,
        plot_all_pmts,
        log=log,
    )

    event = np.zeros(1, dtype=strax.time_fields)
    event["time"] = peaks[0]["time"]
    event["endtime"] = strax.endtime(peaks)[-1]
    waveform = plot_event(peaks, signal, labels, event[0], colors, yscale[-1])

    # Create tile:
    title = _make_peaks_title(peaks, run_id)

    # Put everything together:
    if bottom_pmt_array:
        upper_row = [fig_s1, fig_s2, fig_top, fig_bottom]
    else:
        upper_row = [fig_s1, fig_s2, fig_top]

    upper_row = bokeh.layouts.Row(
        children=upper_row,
        sizing_mode="scale_width",
    )

    plots = bokeh.layouts.gridplot(
        children=[upper_row, waveform],
        sizing_mode="scale_width",
        ncols=1,
        merge_tools=True,
        toolbar_location="above",
    )
    event_display = bokeh.layouts.Column(
        children=[title, plots],
        sizing_mode="scale_width",
        max_width=1600,
    )

    if _provide_peaks:
        return event_display, peaks

    return event_display


def plot_peak_detail(
    peak,
    time_scalar=1,
    label="",
    unit="ns",
    colors=("gray", "blue", "green"),
    fig=None,
):
    """Function which makes a detailed plot for the given peak. As in the main/alt S1/S2 plots of
    the event display.

    :param peak: Peak to be plotted.
    :param time_scalar: Factor to rescale the time from ns to other scale. E.g. =1000 scales to µs.
    :param label: Label to be used in the plot legend.
    :param unit: Time unit of the plotted peak.
    :param colors: Colors to be used for unknown, s1 and s2 peaks.
    :param fig: Instance of bokeh.plotting.figure if None one will be created via
        straxen.bokeh.utils.default_figure().
    :return: Instance of bokeh.plotting.figure

    """
    if not peak.shape:
        peak = np.array([peak])

    if peak.shape[0] != 1:
        raise ValueError(
            "Cannot plot the peak details for more than one "
            "peak. Please make sure peaks has the shape (1,)!"
        )

    p_type = peak[0]["type"]

    if not fig:
        fig = straxen.bokeh_utils.default_fig(title=f"Main/Alt S{p_type}")

    source = straxen.bokeh_utils.get_peaks_source(
        peak,
        relative_start=peak[0]["time"],
        time_scaler=time_scalar,
        keep_amplitude_per_sample=False,
    )

    _i = p_type if p_type < len(colors) else 0
    patches = fig.patches(
        source=source,
        legend_label=label,
        fill_color=colors[_i],
        fill_alpha=0.2,
        line_color=colors[_i],
        line_width=0.5,
        name=label,
    )
    tt = straxen.bokeh_utils.peak_tool_tip(p_type)
    tt = [v for k, v in tt.items() if k not in ["time_static", "center_time", "endtime"]]
    fig.add_tools(bokeh.models.HoverTool(name=label, tooltips=tt, renderers=[patches]))
    fig.xaxis.axis_label = f"Time [{unit}]"
    fig.xaxis.axis_label_text_font_size = "12pt"
    fig.yaxis.axis_label = "Amplitude [pe/ns]"
    fig.yaxis.axis_label_text_font_size = "12pt"

    fig.legend.location = "top_right"
    fig.legend.click_policy = "hide"

    if label:
        fig.legend.visible = True
    else:
        fig.legend.visible = False

    return fig, patches


def plot_peaks(peaks, time_scalar=1, fig=None, colors=("gray", "blue", "green"), yscale="linear"):
    """Function which plots a list/array of peaks relative to the first one.

    :param peaks: Peaks to be plotted.
    :param time_scalar: Factor to rescale the time from ns to other scale. E.g. =1000 scales to µs.
    :param colors: Colors to be used for unknown, s1 and s2 signals
    :param yscale: yscale type can be "linear" or "log"
    :param fig: Instance of bokeh.plotting.figure if None one will be created via
        straxen.bokeh.utils.default_figure().
    :return: bokeh.plotting.figure instance.

    """
    if not fig:
        fig = straxen.bokeh_utils.default_fig(width=1600, height=400, y_axis_type=yscale)

    for i in np.unique(peaks["type"]):
        _ind = np.where(peaks["type"] == i)[0]
        if not len(_ind):
            continue

        source = straxen.bokeh_utils.get_peaks_source(
            peaks[_ind],
            relative_start=peaks[0]["time"],
            time_scaler=time_scalar,
            keep_amplitude_per_sample=False,
        )

        _i = i if i < len(colors) else 0
        p = fig.patches(
            source=source,
            fill_color=colors[_i],
            fill_alpha=0.2,
            line_color=colors[_i],
            line_width=0.5,
            legend_label=LEGENDS[_i],
            name=LEGENDS[_i],
        )

        tt = straxen.bokeh_utils.peak_tool_tip(i)
        tt = [v for k, v in tt.items() if k != "time_dynamic"]
        fig.add_tools(bokeh.models.HoverTool(name=LEGENDS[_i], tooltips=tt, renderers=[p]))
        fig.add_tools(bokeh.models.WheelZoomTool(dimensions="width", name="wheel"))
        fig.toolbar.active_scroll = [t for t in fig.tools if t.name == "wheel"][0]

    fig.xaxis.axis_label = "Time [µs]"
    fig.xaxis.axis_label_text_font_size = "12pt"
    fig.yaxis.axis_label = "Amplitude [pe/ns]"
    fig.yaxis.axis_label_text_font_size = "12pt"

    fig.legend.location = "top_left"
    fig.legend.click_policy = "hide"
    return fig


def plot_pmt_array(
    peak,
    array_type,
    to_pe,
    plot_all_pmts=False,
    log=False,
    fig=None,
    label="",
):
    """Plots top or bottom PMT array for given peak.

    :param peak: Peak for which the hit pattern should be plotted.
    :param array_type: String which specifies if "top" or "bottom" PMT array should be plotted
    :param to_pe: PMT gains.
    :param log: If true use a log-scale for the color scale.
    :param plot_all_pmts: If True colors all PMTs instead of showing swtiched off PMTs as gray dots.
    :param fig: Instance of bokeh.plotting.figure if None one will be created via
        straxen.bokeh.utils.default_figure().
    :param label: Label of the peak which should be used for the plot legend
    :return: Tuple containing a bokeh figure, glyph and transform instance.

    """
    if peak.shape:
        raise ValueError("Can plot PMT array only for a single peak at a time.")

    tool_tip = [
        ("plot", "$name"),
        ("channel", "@pmt"),
        ("x [cm]", "$x"),
        ("y [cm]", "$y"),
        ("area [pe]", "@area"),
    ]

    array = ("top", "bottom")
    if array_type not in array:
        raise ValueError('"array_type" must be either top or bottom.')

    if not fig:
        fig = straxen.bokeh_utils.default_fig(title=f"{array_type} array")

    # Creating TPC axis and title
    fig = _plot_tpc(fig)

    # Plotting PMTs:
    pmts = straxen.pmt_positions()
    if plot_all_pmts:
        mask_pmts = np.zeros(len(pmts), dtype=bool)
    else:
        mask_pmts = to_pe == 0
    pmts_on = pmts[~mask_pmts]
    pmts_on = pmts_on[pmts_on["array"] == array_type]

    if np.any(mask_pmts):
        pmts_off = pmts[mask_pmts]
        pmts_off = pmts_off[pmts_off["array"] == array_type]
        fig = _plot_off_pmts(pmts_off, fig)

    area_per_channel = peak["area_per_channel"][pmts_on["i"]]

    if log:
        area_plot = np.log10(area_per_channel)
        # Manually set infs to zero since cmap cannot handle it.
        area_plot = np.where(area_plot == -np.inf, 0, area_plot)
    else:
        area_plot = area_per_channel

    mapper = bokeh.transform.linear_cmap(
        field_name="area_plot", palette="Viridis256", low=min(area_plot), high=max(area_plot)
    )

    source_on = bklt.ColumnDataSource(
        data={
            "x": pmts_on["x"],
            "y": pmts_on["y"],
            "area": area_per_channel,
            "area_plot": area_plot,
            "pmt": pmts_on["i"],
        }
    )

    p = fig.scatter(
        source=source_on,
        radius=straxen.tpc_pmt_radius,
        fill_color=mapper,
        fill_alpha=1,
        line_color="black",
        legend_label=label,
        name=label + " PMT array",
    )
    fig.add_tools(
        bokeh.models.HoverTool(name=label + " PMT array", tooltips=tool_tip, renderers=[p])
    )
    fig.legend.location = "top_left"
    fig.legend.click_policy = "hide"
    fig.legend.orientation = "horizontal"
    fig.legend.padding = 0
    fig.toolbar_location = None
    return fig, p, mapper


def _plot_tpc(fig=None):
    """Plots ring at TPC radius and sets xy limits + labels."""
    if not fig:
        fig = straxen.bokeh_utils.default_fig()

    fig.circle(
        x=0,
        y=0,
        radius=straxen.tpc_r,
        fill_color="white",
        line_color="black",
        line_width=3,
        fill_alpha=0,
    )
    fig.xaxis.axis_label = "x [cm]"
    fig.xaxis.axis_label_text_font_size = "12pt"
    fig.yaxis.axis_label = "y [cm]"
    fig.yaxis.axis_label_text_font_size = "12pt"
    fig.x_range.start = -80
    fig.x_range.end = 80
    fig.y_range.start = -80
    fig.y_range.end = 80

    return fig


def _plot_off_pmts(pmts, fig=None):
    """Plots PMTs which are switched off."""
    if not fig:
        fig = straxen.bokeh_utils.default_fig()
    fig.circle(
        x=pmts["x"],
        y=pmts["y"],
        fill_color="gray",
        line_color="black",
        radius=straxen.tpc_pmt_radius,
    )
    return fig


def plot_posS2s(peaks, label="", fig=None, s2_type_style_id=0):
    """Plots xy-positions of specified peaks.

    :param peaks: Peaks for which the position should be plotted.
    :param label: Legend label and plot name (name serves as idenitfier).
    :param fig: bokeh.plotting.figure instance the plot should be plotted into. If None creates new
        instance.
    :param s2_type_style_id: 0 plots main S2 style, 1 for alt S2 and 2 for other S2s (e.g. single
        electrons).

    """
    if not peaks.shape:
        peaks = np.array([peaks])

    if not fig:
        fig = straxen.bokeh_utils.default_fig()

    source = straxen.bokeh_utils.get_peaks_source(peaks)

    if s2_type_style_id == 0:
        color = "red"
        p = fig.cross(
            source=source, name=label, legend_label=label, color=color, line_width=2, size=12
        )
    elif s2_type_style_id == 1:
        color = "orange"
        p = fig.cross(
            source=source,
            name=label,
            legend_label=label,
            color=color,
            angle=45 / 360 * 2 * np.pi,
            line_width=2,
            size=12,
        )
    else:
        color = "red"
        p = fig.diamond_cross(source=source, name=label, legend_label=label, color=color, size=8)

    tt = straxen.bokeh_utils.peak_tool_tip(2)
    tt = [v for k, v in tt.items() if k not in ["time_dynamic", "amplitude"]]
    fig.add_tools(
        bokeh.models.HoverTool(
            name=label,
            tooltips=[("x [cm]", "@x"), ("y [cm]", "@y")] + tt,
            renderers=[p],
        )
    )

    if "position_contour_cnf" in peaks.dtype.names:
        # Plotting the contour of the S2
        c = fig.multi_line(
            peaks["position_contour_cnf"][:, :, 0].tolist(),
            peaks["position_contour_cnf"][:, :, 1].tolist(),
            name=label,
            legend_label=label,
            color=color,
            line_width=1.0,
        )
        return fig, [p, c]
    return fig, [p]


def _make_event_title(event, run_id, width=1600):
    """Function which makes the title of the plot for the specified event.

    Note:
        To center the title I use a transparent box.

    :param event: Event which we are plotting
    :param run_id: run_id

    :return: Title as bokeh.models.Div instance

    """
    start = event["time"]
    date = np.datetime_as_string(start.astype("<M8[ns]"), unit="s")
    start_ns = start - (start // straxen.units.s) * straxen.units.s
    end = strax.endtime(event)
    end_ns = end - start + start_ns
    event_number = event["event_number"]
    text = (
        f"<h2>Event {event_number} from run {run_id}<br>"
        f"Recorded at {date[:10]} {date[10:]} UTC, {start_ns} ns - {end_ns} ns<br>"
        f"({start} - {end})</h2>"
    )

    title = bokeh.models.Div(
        text=text,
        styles={
            "text-align": "left",
        },
        sizing_mode="scale_width",
        width=width,
        # orientation='vertical',
        width_policy="fit",
        margin=(0, 0, -30, 50),
    )
    return title


def _make_peaks_title(peaks, run_id, width=1600):
    """Function which makes the title of the plot for the specified peaks.

    Note:
        To center the title I use a transparent box.

    :param peaks: Peaks which we are plotting
    :param run_id: run_id

    :return: Title as bokeh.models.Div instance

    """
    start = peaks["time"].min()
    date = np.datetime_as_string(start.astype("<M8[ns]"), unit="s")
    start_ns = start - (start // straxen.units.s) * straxen.units.s
    end = strax.endtime(peaks).max()
    end_ns = end - start + start_ns
    text = (
        f"<h2>Peaks from run {run_id}<br>"
        f"Recorded at {date[:10]} {date[10:]} UTC, {start_ns} ns - {end_ns} ns<br>"
        f"({start} - {end})</h2>"
    )

    title = bokeh.models.Div(
        text=text,
        styles={
            "text-align": "left",
        },
        sizing_mode="scale_width",
        width=width,
        # orientation='vertical',
        width_policy="fit",
        margin=(0, 0, -30, 50),
    )
    return title


def bokeh_set_x_range(plot, x_range, debug=False):
    """Function which adjust java script call back for x_range of a bokeh plot. Required to link
    bokeh and holoviews x_range.

    Note:
        This is somewhat voodoo + some black magic,
        but it works....

    """
    from bokeh.models import CustomJS

    code = """\
    const start = cb_obj.start;
    const end = cb_obj.end;
    // Need to update the attributes at the same time.
    x_range.setv({start, end});
    """
    for attr in ["start", "end"]:
        if debug:
            # Prints x_range bar to check Id, as I said voodoo
            print(x_range)
        plot.x_range.js_on_change(attr, CustomJS(args=dict(x_range=x_range), code=code))


class DataSelectionHist:
    """Class for an interactive data selection plot."""

    def __init__(self, name, size=600):
        """Class for an interactive data selection plot.

        :param name: Name of the class object instance. Needed for dynamic return, e.g. ds =
            DataSelectionHist("ds")
        :param size: Edge size of the figure in pixel.

        """
        raise NotImplementedError(
            "This function does not work with"
            " the latest bokeh version. If you are still using"
            " this function please let us know in tech-support"
            " by the 01.04.2024, else we reomve this function."
        )
        self.name = name
        self.selection_index = None
        self.size = size

        from bokeh.io import output_notebook

        output_notebook()

    def histogram2d(
        self,
        items,
        xdata,
        ydata,
        bins,
        hist_range,
        x_label="X-Data",
        y_label="Y-Data",
        log_color_scale=True,
        cmap_steps=256,
        clim=(None, None),
        undeflow_color=None,
        overflow_color=None,
        weights=1,
    ):
        """2d Histogram which allows to select the plotted items dynamically.

        Note:
            You can select the data either via a box select or Lasso
            select tool. The data can be returned by:

            ds.get_back_selected_items()

            Hold shift to select multiple regions.

        Warnings:
            Depending on the number of bins the Lasso selection can
            become relatively slow. The number of bins should not be
            larger than 100.
            The box selection performance is better.

        :param items: numpy.structured.array of items to be selected.
            e.g. peaks or events.
        :param xdata: numpy.array for xdata e.g. peaks['area']
        :param ydata: same
        :param bins: Integer specifying the number of bins. Currently
            x and y axis must share the same binning.
        :param hist_range: Tuple of x-range and y-range.
        :param x_label: Label to be used for the x-axis
        :param y_label: same but for y
        :param log_color_scale: If true (default) use log colorscale
        :param cmap_steps: Integer between 0 and 256 for stepped
            colorbar.
        :param clim: Tuple of color limits.
        :param undeflow_color: If specified colors all bins below clim
            with the corresponding color.
        :param overflow_color: Same but per limit.
        :param weights: If specified each bin entry is weighted by this
            value. Can be either a scalar e.g. a time or an array of
            weights which has the same length as the x/y data.
        :return: bokeh figure instance.

        """
        if isinstance(bins, tuple):
            raise ValueError(
                "Currently only squared bins are supported. Plase change bins into an integer."
            )

        x_pos, y_pos = self._make_bin_positions((bins, bins), hist_range)
        weights = np.ones(len(xdata)) * weights

        hist, hist_inds = self._hist2d_with_index(xdata, ydata, weights, self.xedges, self.yedges)

        # Define times and ids for return:
        self.items = items
        self.hist_inds = hist_inds

        colors = self._get_color(
            hist,
            cmap_steps,
            log_color_scale=log_color_scale,
            clim=clim,
            undeflow_color=undeflow_color,
            overflow_color=overflow_color,
        )

        # Create Figure and add LassoTool:
        f = bokeh.plotting.figure(
            title="DataSelection", width=self.size, height=self.size, tools="box_select,reset,save"
        )

        # Add hover tool, colorbar is too complictaed:
        tool_tip = [("Bin Center x", "@x"), ("Bin Center y", "@y"), ("Entries", "@h")]
        f.add_tools(
            bokeh.models.LassoSelectTool(select_every_mousemove=False),
            bokeh.models.HoverTool(tooltips=tool_tip),
        )

        s1 = bokeh.plotting.ColumnDataSource(
            data=dict(x=x_pos, y=y_pos, h=hist.flatten(), color=colors)
        )
        f.square(source=s1, size=self.size / bins, color="color", nonselection_alpha=0.3)

        f.x_range.start = self.xedges[0]
        f.x_range.end = self.xedges[-1]
        f.y_range.start = self.yedges[0]
        f.y_range.end = self.yedges[-1]
        f.xaxis.axis_label = x_label
        f.yaxis.axis_label = y_label

        self.selection_index = None
        s1.selected.js_on_change(
            "indices",
            bokeh.models.CustomJS(
                args=dict(s1=s1),
                code=f"""
                var inds = cb_obj.indices;
                var kernel = IPython.notebook.kernel;
                kernel.execute("{self.name}.selection_index = " + inds);
                """,
            ),
        )
        return f

    def get_back_selected_items(self):
        if not self.selection_index:
            raise ValueError(
                "No data selection found. Have you selected any data? "
                "If yes you most likely have not intialized the DataSelctor correctly. "
                'You have to callit as: my_instance_name = DataSelectionHist("my_instance_name")'
            )
        m = np.isin(self.hist_inds, self.selection_index)
        return self.items[m]

    @staticmethod
    @numba.njit
    def _hist2d_with_index(xdata, ydata, weights, x_edges, y_edges):
        n_x_bins = len(x_edges) - 1
        n_y_bins = len(y_edges) - 1
        res_hist_inds = np.zeros(len(xdata), dtype=np.int32)
        res_hist = np.zeros((n_x_bins, n_y_bins), dtype=np.int64)

        # Create bin ranges:
        offset = 0
        for ind, xv in enumerate(xdata):
            yv = ydata[ind]
            w = weights[ind]
            hist_ind = 0
            found = False
            for ind_xb, low_xb in enumerate(x_edges[:-1]):
                high_xb = x_edges[ind_xb + 1]

                if not low_xb <= xv:
                    hist_ind += n_y_bins
                    continue
                if not xv < high_xb:
                    hist_ind += n_y_bins
                    continue

                # Checked both bins value is in bin, so check y:
                for ind_yb, low_yb in enumerate(y_edges[:-1]):
                    high_yb = y_edges[ind_yb + 1]

                    if not low_yb <= yv:
                        hist_ind += 1
                        continue
                    if not yv < high_yb:
                        hist_ind += 1
                        continue

                    found = True
                    res_hist_inds[offset] = hist_ind
                    res_hist[ind_xb, ind_yb] += w
                    offset += 1

            # Set to -1 if not in any
            if not found:
                res_hist_inds[offset] = -1
                offset += 1
        return res_hist, res_hist_inds

    def _make_bin_positions(self, bins, bin_range):
        """Helper function to create center positions for "histogram" markers."""
        edges = []
        for b, br in zip(bins, bin_range):
            # Create x and y edges
            d_range = br[1] - br[0]
            edges.append(np.arange(br[0], br[1] + d_range / b, d_range / b))

        # Convert into marker positions:
        xedges = edges[0]
        yedges = edges[1]
        self.xedges = xedges
        self.yedges = yedges
        x_pos = xedges[:-1] + np.diff(xedges) / 2
        x_pos = np.repeat(x_pos, len(yedges) - 1)

        y_pos = yedges[:-1] + np.diff(yedges) / 2
        y_pos = np.array(list(y_pos) * (len(xedges) - 1))
        return x_pos, y_pos

    def _get_color(
        self,
        hist,
        cmap_steps,
        log_color_scale=False,
        clim=(None, None),
        undeflow_color=None,
        overflow_color=None,
    ):
        """Helper function to create colorscale."""
        hist = hist.flatten()

        if clim[0] and undeflow_color:
            # If underflow is specified get indicies for underflow bins
            inds_underflow = np.argwhere(hist < clim[0]).flatten()

        if clim[1] and overflow_color:
            inds_overflow = np.argwhere(hist > clim[1]).flatten()

        # Clip data according to clim
        if np.any(clim):
            hist = np.clip(hist, clim[0], clim[1])

        self.clim = (np.min(hist), np.max(hist))
        if log_color_scale:
            color = np.log10(hist)
            color /= np.max(color)
            color *= cmap_steps - 1
        else:
            color = hist / np.max(hist)
            color *= cmap_steps - 1

        cmap = np.array(bokeh.palettes.viridis(cmap_steps))
        cmap = cmap[np.round(color).astype(np.int8)]

        if undeflow_color:
            cmap[inds_underflow] = undeflow_color

        if overflow_color:
            cmap[inds_overflow] = overflow_color
        return cmap



import bokeh.plotting as bklt
from bokeh.embed import file_html
from bokeh.resources import CDN
import numpy as np
import strax

export, __all__ = strax.exporter()


@export
def bokeh_to_wiki(fig, outputfile=None):
    """Function which converts bokeh HTML code to a wiki readable code.

    :param fig: Figure to be conerted
    :param outputfile: String of absolute file path. If specified output is writen to the file. Else
        output is print to the notebook and can be simply copied into the wiki.

    """
    # convert plot to wiki format:
    html = file_html(fig, CDN)
    html = "\n".join((["<html>"] + html.split("\n")[2:]))

    if outputfile:
        with open(outputfile, mode="w") as file:
            file.write(html)
    else:
        print(html)


def get_peaks_source(peaks, relative_start=0, time_scaler=1, keep_amplitude_per_sample=True):
    """Computes bokeh.plotting.ColumnDataSource for given peaks.

    :param peaks: Peaks to be plotted.
    :param relative_start: t0 from which on the peaks should be plotted.
    :param time_scaler: Factor to rescale the time from ns to other scale. E.g. =1000 scales to µs.
    :param keep_amplitude_per_sample: Boolean if true amplitude of the plotted peaks is in
        pe/sample. False pe/ns.
    :return: bokeh.plotting.ColumnDataSource instance which can be used to plot peaks.

    """
    if not (np.all(peaks["type"] == peaks[0]["type"])):
        raise ValueError("All peaks must be of the same type (S1, S2 or Unknown)!")

    x_p = []
    y_p = []
    for ind, p in enumerate(peaks):
        x, y = _patches_x_y(p, keep_amplitude_per_sample=keep_amplitude_per_sample)
        x -= relative_start  # relative to first peak
        x = x / time_scaler
        x_p.append(x)
        y_p.append(y)

    if peaks[0]["type"] == 2:
        scaler = 10**-3
    else:
        scaler = 1

    source = bklt.ColumnDataSource(
        data={
            "type": peaks["type"],
            "xs": x_p,  # Coordinates for peak patches
            "ys": y_p,
            "x": peaks["x"],  # XY-Pos in PMZ Hitpattern
            "y": peaks["y"],
            "dt": peaks["dt"],
            "time": peaks["time"],
            "center_time": peaks["center_time"],
            "endtime": strax.endtime(peaks),
            "area": peaks["area"],
            "nhits": peaks["n_hits"],
            "aft": peaks["area_fraction_top"],
            "width_50": peaks["range_50p_area"] * scaler,
            "width_90": peaks["range_90p_area"] * scaler,
            "rise": peaks["rise_time"] * scaler,
            "rel_center_time": (peaks["center_time"] - peaks["time"]) * scaler,
        }
    )
    return source


def _patches_x_y(peak, keep_amplitude_per_sample=False):
    """Creates x,y coordinates needed to draw peaks via bokeh.models.patches.

    :param peak: Peak for which we need the x/y samples
    :param keep_amplitude_per_sample: If y-data should be in units of "per sample" or "per ns".
    :return: Tuple of x, y

    """
    if keep_amplitude_per_sample:
        dt_a = 1
    else:
        dt_a = peak["dt"]

    x = np.arange(peak["length"])
    xx = np.zeros(2 * len(x), dtype=x.dtype)
    mx = 0.5 * (x[1::] + x[:-1])
    xx[1:-1:2] = mx
    xx[2::2] = mx
    xx[0] = 1.5 * x[0] - 0.5 * x[1]
    xx[-1] = 1.5 * x[-1] - 0.5 * x[-2]
    xx = np.array(xx) * peak["dt"] + peak["time"]
    y = peak["data"][: peak["length"]]
    yy = np.zeros(2 * len(x))
    yy[0::2] = y
    yy[1::2] = y
    yy = np.array(yy) / dt_a

    # baseline since we'll fill underneath
    xx = np.concatenate([[xx[0]], xx, [xx[-1]]])
    yy = np.concatenate([[0], yy, [0]])
    return xx, yy


def peak_tool_tip(peak_type):
    """Default mouseover tooltip for peaks.

    :param peak_type: If 2, all time variables are in µs else in ns.
    :return: dictionary of tooltips. Can be converted to a list for bokeh.models.HoverTool.

    """

    tool_tip = dict()
    tool_tip["type"] = ("type", "@type")

    # Add non-time parameters (results in an ordered tooltip)
    tool_tip["amplitude"] = ("amplitude [pe/ns]", "$y")
    tool_tip["area"] = ("area [pe]", "@area")
    tool_tip["x"] = ("x [cm]", "@x")
    tool_tip["y"] = ("y [cm]", "@y")

    # Now ns/µs parameters for S1 and S2
    tool_tip["time_dynamic"] = ("time [ns]", "$x")
    tool_tip["dt"] = ("dt [ns/sample]", "@dt")
    tool_tip["range_50p_width"] = ("50% width [ns]", "@width_50")
    tool_tip["range_90p_width"] = ("90% width [ns]", "@width_90")
    tool_tip["rise_time"] = ("rise time [ns]", "@rise")
    tool_tip["rel_center_time"] = ("center time [ns]", "@rel_center_time")

    # Misc parameters
    tool_tip["nhits"] = ("nhits", "@nhits")
    tool_tip["aft"] = ("AFT", "@aft")

    # Add static time parameters:
    tool_tip["time_static"] = ("time [ns]", "@time")
    tool_tip["center_time"] = ("center_time [ns]", "@center_time")
    tool_tip["endtime"] = ("endtime [ns]", "@endtime")

    if peak_type == 2:
        for k, i in tool_tip.items():
            if k not in ["time_static", "center_time", "endtime"]:
                tool_tip[k] = (i[0].replace("[ns]", "[µs]"), i[1])

    return tool_tip


def default_fig(width=400, height=400, title="", **kwargs):
    """Helper function which returns a bokeh.plotting.figure instance with sizing_mode set to
    'scale_both' and an aspect ratio set according to the specified width and height.

    :param width: Plot width in pixels
    :param height: PLot height in pixels.
    :param title: Title of the plot. Also allows for additional kwargs accepted by bokeh.plotting.
    :return: bokeh.plotting.figure instance.

    """
    fig = bklt.figure(
        width=width,
        height=height,
        sizing_mode="scale_width",
        title=title,
        tools="pan,box_zoom,reset,save",
        **kwargs,
    )
    return fig



from typing import Dict, Tuple, Union
import numba
import numpy as np
from immutabledict import immutabledict
import strax
from strax.processing.general import _touching_windows
from strax.dtypes import DIGITAL_SUM_WAVEFORM_CHANNEL
import straxen


export, __all__ = strax.exporter()


@export
class Peaklets(strax.Plugin):
    """
    Split records into:
     - peaklets
     - lone_hits

    Peaklets are very aggressively split peaks such that we are able
    to find S1-S2s even if they are close to each other. (S2) Peaks
    that are split into too many peaklets will be merged later on.

    To get Peaklets from records apply/do:
        1. Hit finding
        2. Peak finding
        3. Peak splitting using the natural breaks algorithm
        4. Compute the digital sum waveform

    Lone hits are all hits which are outside of any peak. The area of
    lone_hits includes the left and right hit extension, except the
    extension overlaps with any peaks or other hits.
    """

    depends_on = "records"
    provides: Union[Tuple[str, ...], str] = ("peaklets", "lone_hits")
    data_kind: Union[Dict[str, str], str] = dict(peaklets="peaklets", lone_hits="lone_hits")
    parallel = "process"
    compressor = "zstd"

    rechunk_on_load = True
    chunk_source_size_mb = 100

    # To reduce the number of chunks, we increase the target size
    # This would not harm memory usage, because we rechunk on load
    chunk_target_size_mb = 2000

    __version__ = "1.2.2"

    peaklet_gap_threshold = straxen.URLConfig(
        default=700, infer_type=False, help="No hits for this many ns triggers a new peak"
    )

    peak_left_extension = straxen.URLConfig(
        default=30, infer_type=False, help="Include this many ns left of hits in peaks"
    )

    peak_right_extension = straxen.URLConfig(
        default=200, infer_type=False, help="Include this many ns right of hits in peaks"
    )

    peak_min_pmts = straxen.URLConfig(
        default=2,
        infer_type=False,
        help="Minimum number of contributing PMTs needed to define a peak",
    )

    peak_split_gof_threshold = straxen.URLConfig(
        # See https://xe1t-wiki.lngs.infn.it/doku.php?id=
        # xenon:xenonnt:analysis:strax_clustering_classification
        # #natural_breaks_splitting
        # for more information
        default=(None, ((0.5, 1.0), (6.0, 0.4)), ((2.5, 1.0), (5.625, 0.4))),  # Reserved
        infer_type=False,
        help=(
            "Natural breaks goodness of fit/split threshold to split "
            "a peak. Specify as tuples of (log10(area), threshold)."
        ),
    )

    peak_split_filter_wing_width = straxen.URLConfig(
        default=70,
        infer_type=False,
        help="Wing width of moving average filter for low-split natural breaks",
    )

    peak_split_min_area = straxen.URLConfig(
        default=40.0,
        infer_type=False,
        help="Minimum area to evaluate natural breaks criterion. Smaller peaks are not split.",
    )

    peak_split_iterations = straxen.URLConfig(
        default=20, infer_type=False, help="Maximum number of recursive peak splits to do."
    )

    diagnose_sorting = straxen.URLConfig(
        track=False,
        default=False,
        infer_type=False,
        help="Enable runtime checks for sorting and disjointness",
    )

    gain_model = straxen.URLConfig(
        infer_type=False, help="PMT gain model. Specify as URL or explicit value"
    )

    tight_coincidence_window_left = straxen.URLConfig(
        default=50,
        infer_type=False,
        help="Time range left of peak center to call a hit a tight coincidence (ns)",
    )

    tight_coincidence_window_right = straxen.URLConfig(
        default=50,
        infer_type=False,
        help="Time range right of peak center to call a hit a tight coincidence (ns)",
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    n_top_pmts = straxen.URLConfig(type=int, help="Number of top TPC array PMTs")

    store_data_top = straxen.URLConfig(
        default=True, type=bool, help="Save the sum waveform of the top array separately"
    )

    store_data_start = straxen.URLConfig(
        default=True, type=bool, help="Save the start time of the waveform with minimum dt"
    )

    saturation_correction_on = straxen.URLConfig(
        default=True, infer_type=False, help="On off switch for saturation correction"
    )

    saturation_reference_length = straxen.URLConfig(
        default=100,
        infer_type=False,
        help="Maximum number of reference sample used to correct saturated samples",
    )

    saturation_min_reference_length = straxen.URLConfig(
        default=20,
        infer_type=False,
        help="Minimum number of reference sample used to correct saturated samples",
    )

    peaklet_max_duration = straxen.URLConfig(
        default=int(10e6), infer_type=False, help="Maximum duration [ns] of a peaklet"
    )

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number.",
    )

    hit_min_amplitude = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://hit_thresholds"
            "?as_list=True"
            "&sort=pmt"
            "&attr=value"
            "&detector=tpc"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
        ),
        help="Minimum hit amplitude in ADC counts above baseline. "
        "Specify as a tuple of length n_tpc_pmts, or a number,"
        'or a string like "pmt_commissioning_initial" which means calling'
        "hitfinder_thresholds.py",
        track=True,
        infer_type=False,
    )

    def infer_dtype(self):
        return dict(
            peaklets=strax.peak_dtype(
                n_channels=self.n_tpc_pmts,
                store_data_top=self.store_data_top,
                store_data_start=self.store_data_start,
            ),
            lone_hits=strax.hit_dtype,
        )

    def setup(self):
        if self.peak_min_pmts > 2:
            # Can fix by re-splitting,
            raise NotImplementedError(
                f"Raising the peak_min_pmts to {self.peak_min_pmts} "
                "interferes with lone_hit definition. "
                "See github.com/XENONnT/straxen/issues/295"
            )

        self.to_pe = self.gain_model

        self.hit_thresholds = self.hit_min_amplitude

        self.channel_range = self.channel_map["tpc"]

        self._tight_coincidence_window_left = self.tight_coincidence_window_left
        self._tight_coincidence_window_right = self.tight_coincidence_window_right

        if self.peaklet_gap_threshold > strax.DEFAULT_CHUNK_SPLIT_NS:  # 1000 ns
            raise ValueError(
                f"peaklet_gap_threshold {self.peaklet_gap_threshold} ns "
                "in peaklets building is larger than "
                f"safe_break_in_pulses {strax.DEFAULT_CHUNK_SPLIT_NS} ns "
                "in raw_records building. "
                "This is inconsistent because raw_records can not be split "
                "if the nearby hits are too close."
            )

    def compute(self, records, start, end):
        hits = strax.find_hits(records, min_amplitude=self.hit_thresholds)

        # Remove hits in zero-gain channels
        # they should not affect the clustering!
        hits = hits[self.to_pe[hits["channel"]] != 0]

        hits = strax.sort_by_time(hits)

        # Use peaklet gap threshold for initial clustering
        # based on gaps between hits
        peaklets = strax.find_peaks(
            hits,
            self.to_pe,
            gap_threshold=self.peaklet_gap_threshold,
            left_extension=self.peak_left_extension,
            right_extension=self.peak_right_extension,
            min_channels=self.peak_min_pmts,
            # NB, need to have the data_top field here, will discard if not digitized later
            result_dtype=strax.peak_dtype(
                n_channels=self.n_tpc_pmts, store_data_top=True, store_data_start=True
            ),
            max_duration=self.peaklet_max_duration,
        )

        # Make sure peaklets don't extend out of the chunk boundary
        # This should be very rare in normal data due to the ADC pretrigger
        # window.
        # The different chunking causes different results, but this is
        # NOT tracked by lineage
        self.clip_peaklet_times(peaklets, start, end)

        # Get hits outside peaklets, and store them separately.
        # fully_contained is OK provided gap_threshold > extension,
        # which is asserted inside strax.find_peaks.
        is_lone_hit = strax.fully_contained_in(hits, peaklets) == -1
        lone_hits = hits[is_lone_hit]
        strax.integrate_lone_hits(
            lone_hits,
            records,
            peaklets,
            save_outside_hits=(self.peak_left_extension, self.peak_right_extension),
            n_channels=len(self.to_pe),
        )
        if np.any(lone_hits["right_integration"] - lone_hits["left_integration"] <= 0):
            raise ValueError("Find lone_hits with non-positive length!")

        # Compute basic peak properties -- needed before natural breaks
        hits = hits[~is_lone_hit]
        # Define regions outside of peaks such that _find_hit_integration_bounds
        # is not extended beyond a peak.
        outside_peaks = self.create_outside_peaks_region(peaklets, start, end)
        strax.find_hit_integration_bounds(
            hits,
            outside_peaks,
            records,
            save_outside_hits=(self.peak_left_extension, self.peak_right_extension),
            n_channels=len(self.to_pe),
            allow_bounds_beyond_records=True,
        )

        # Transform hits to hitlets for naming conventions. A hit refers
        # to the central part above threshold a hitlet to the entire signal
        # including the left and right extension.
        # (We are not going to use the actual hitlet data_type here.)
        hitlets = hits
        # This line will not clean the memory, but only prevent misinterpretation
        # only if sys.getrefcount(hits) - 1 is 1, we can clean the memory
        del hits

        # Extend hits into hitlets and clip at chunk boundaries:
        hitlets["time"] -= (hitlets["left"] - hitlets["left_integration"]) * hitlets["dt"]
        hitlets["length"] = hitlets["right_integration"] - hitlets["left_integration"]

        hitlets = strax.sort_by_time(hitlets)
        self.clip_peaklet_times(hitlets, start, end)
        rlinks = strax.record_links(records)

        strax.sum_waveform(
            peaklets,
            hitlets,
            records,
            rlinks,
            self.to_pe,
            n_top_channels=self.n_top_pmts,
            store_data_top=self.store_data_top,
            store_data_start=self.store_data_start,
        )

        strax.compute_properties(peaklets, n_top_channels=self.n_top_pmts)

        # Split peaks using low-split natural breaks;
        # see https://github.com/XENONnT/straxen/pull/45
        # and https://github.com/AxFoundation/strax/pull/225
        peaklets = strax.split_peaks(
            peaklets,
            hitlets,
            records,
            rlinks,
            self.to_pe,
            algorithm="natural_breaks",
            threshold=self.natural_breaks_threshold,
            split_low=True,
            filter_wing_width=self.peak_split_filter_wing_width,
            min_area=self.peak_split_min_area,
            do_iterations=self.peak_split_iterations,
            n_top_channels=self.n_top_pmts,
            store_data_top=self.store_data_top,
            store_data_start=self.store_data_start,
        )

        # Saturation correction using non-saturated channels
        # similar method used in pax
        # see https://github.com/XENON1T/pax/pull/712

        if self.saturation_correction_on:
            peak_list = peak_saturation_correction(
                records,
                rlinks,
                peaklets,
                hitlets,
                self.to_pe,
                reference_length=self.saturation_reference_length,
                min_reference_length=self.saturation_min_reference_length,
                n_top_channels=self.n_top_pmts,
                store_data_top=self.store_data_top,
                store_data_start=self.store_data_start,
            )

            # Compute the width again for corrected peaks
            strax.compute_properties(
                peaklets, n_top_channels=self.n_top_pmts, select_peaks_indices=peak_list
            )

        # Compute tight coincidence level.
        # Making this a separate plugin would
        # (a) doing hitfinding yet again (or storing hits)
        # (b) increase strax memory usage / max_messages,
        #     possibly due to its currently primitive scheduling.
        hit_max_times_argsort = strax.stable_argsort(hitlets["max_time"])
        sorted_hit_max_times = hitlets["max_time"][hit_max_times_argsort]
        sorted_hit_channels = hitlets["channel"][hit_max_times_argsort]
        peaklet_max_times = peaklets["time"] + np.argmax(peaklets["data"], axis=1) * peaklets["dt"]
        peaklets["tight_coincidence"] = get_tight_coin(
            sorted_hit_max_times,
            sorted_hit_channels,
            peaklet_max_times,
            self._tight_coincidence_window_left,
            self._tight_coincidence_window_right,
            self.channel_range,
        )

        # Add max and min time difference between apexes of hits
        self.add_hit_features(hitlets, peaklets)

        if self.diagnose_sorting and len(records):
            assert np.diff(records["time"]).min(initial=1) >= 0, "Records not sorted"
            assert np.diff(hitlets["time"]).min(initial=1) >= 0, "Hits/Hitlets not sorted"
            assert np.all(
                peaklets["time"][1:] >= strax.endtime(peaklets)[:-1]
            ), "Peaks not disjoint"

        # Update nhits of peaklets:
        counts = strax.touching_windows(hitlets, peaklets)
        counts = np.diff(counts, axis=1).flatten()
        peaklets["n_hits"] = counts

        # Drop the data_top or data_start field
        if (not self.store_data_top) or (not self.store_data_start):
            peaklets = drop_data_field(peaklets, self.dtype_for("peaklets"))

        # Check channel of peaklets
        peaklets_unique_channel = np.unique(peaklets["channel"])
        if np.any(peaklets_unique_channel != DIGITAL_SUM_WAVEFORM_CHANNEL):
            raise ValueError(
                f"Found channel number of peaklets other than "
                f"{DIGITAL_SUM_WAVEFORM_CHANNEL}: {peaklets_unique_channel.tolist()}"
            )
        # Check tight_coincidence
        if not np.all(peaklets["n_hits"] >= peaklets["tight_coincidence"]):
            raise ValueError(f"Found n_hits less than tight_coincidence")

        return dict(peaklets=peaklets, lone_hits=lone_hits)

    def natural_breaks_threshold(self, peaks):
        rise_time = -peaks["area_decile_from_midpoint"][:, 1]

        # This is ~1 for an clean S2, ~0 for a clean S1,
        # and transitions gradually in between.
        f_s2 = 8 * np.log10(rise_time.clip(1, 1e5) / 100)
        f_s2 = 1 / (1 + np.exp(-f_s2))

        log_area = np.log10(peaks["area"].clip(1, 1e7))
        thresholds = self.peak_split_gof_threshold
        return f_s2 * np.interp(log_area, *np.transpose(thresholds[2])) + (1 - f_s2) * np.interp(
            log_area, *np.transpose(thresholds[1])
        )

    @staticmethod
    @numba.njit(nogil=True, cache=True)
    def clip_peaklet_times(peaklets, start, end):
        for p in peaklets:
            if p["time"] < start:
                p["time"] = start
            if strax.endtime(p) > end:
                p["length"] = (end - p["time"]) // p["dt"]

    @staticmethod
    def create_outside_peaks_region(peaklets, start, end):
        """Creates time intervals which are outside peaks.

        :param peaklets: Peaklets for which intervals should be computed.
        :param start: Chunk start
        :param end: Chunk end
        :return: array of strax.time_fields dtype.

        """
        if not len(peaklets):
            return np.zeros(0, dtype=strax.time_fields)

        outside_peaks = np.zeros(len(peaklets) + 1, dtype=strax.time_fields)

        outside_peaks[0]["time"] = start
        outside_peaks[0]["endtime"] = peaklets[0]["time"]
        outside_peaks[1:-1]["time"] = strax.endtime(peaklets[:-1])
        outside_peaks[1:-1]["endtime"] = peaklets["time"][1:]
        outside_peaks[-1]["time"] = strax.endtime(peaklets[-1])
        outside_peaks[-1]["endtime"] = end
        return outside_peaks

    @staticmethod
    def add_hit_features(hitlets, peaklets):
        """Create hits timing features."""
        peaklets["max_diff"] = -1
        peaklets["min_diff"] = -1
        peaklets["first_channel"] = DIGITAL_SUM_WAVEFORM_CHANNEL
        peaklets["last_channel"] = DIGITAL_SUM_WAVEFORM_CHANNEL
        split_hits = strax.split_by_containment(hitlets, peaklets)
        for peaklet, _hitlets in zip(peaklets, split_hits):
            if len(_hitlets) == 0:
                continue
            argsort = strax.stable_argsort(_hitlets["max_time"])
            sorted_hitlets = _hitlets[argsort]
            time_diff = np.diff(sorted_hitlets["max_time"])
            if len(time_diff) > 0:
                peaklet["max_diff"] = time_diff.max()
                peaklet["min_diff"] = time_diff.min()
            else:
                peaklet["max_diff"] = -1
                peaklet["min_diff"] = -1
            peaklet["first_channel"] = sorted_hitlets[0]["channel"]
            peaklet["last_channel"] = sorted_hitlets[-1]["channel"]


def drop_data_field(peaklets, goal_dtype, _name_function="_drop_data_field"):
    """Return peaklets without the data_* field."""
    peaklets_without_field = np.zeros(len(peaklets), dtype=goal_dtype)
    strax.copy_to_buffer(peaklets, peaklets_without_field, _name_function)
    del peaklets
    return peaklets_without_field


@numba.jit(nopython=True, nogil=True, cache=False)
def peak_saturation_correction(
    records,
    rlinks,
    peaks,
    hitlets,
    to_pe,
    reference_length=100,
    min_reference_length=20,
    use_classification=False,
    n_top_channels=0,
    store_data_top=False,
    store_data_start=False,
):
    """Correct the area and per pmt area of peaks from saturation.

    :param records: Records
    :param rlinks: strax.record_links of corresponding records.
    :param peaks: Peaklets / Peaks
    :param hitlets: Hitlets found in records to build peaks. (Hitlets are hits including the
        left/right extension)
    :param to_pe: adc to PE conversion (length should equal number of PMTs)
    :param reference_length: Maximum number of reference sample used to correct saturated samples
    :param min_reference_length: Minimum number of reference sample used to correct saturated
        samples
    :param use_classification: Option of using classification to pick only S2
    :param n_top_channels: Number of top array channels.
    :param store_data_top: Boolean which indicates whether to store the top array waveform in the
        peak.
    :param store_data_start: Boolean which indicates whether to store the first samples of the
        waveform in the peak.

    """

    if not len(records):
        return
    if not len(peaks):
        return

    # Search for peaks with saturated channels
    mask = peaks["n_saturated_channels"] > 0
    if use_classification:
        mask &= peaks["type"] == 2
    peak_list = np.where(mask)[0]
    # Look up records that touch each peak
    record_ranges = _touching_windows(
        records["time"],
        strax.endtime(records),
        peaks[peak_list]["time"],
        strax.endtime(peaks[peak_list]),
    )

    # Create temporary arrays for calculation
    dt = records[0]["dt"]
    n_channels = len(peaks[0]["saturated_channel"])
    len_buffer = np.max(peaks["length"] * peaks["dt"]) // dt + 1
    max_nrecord = len_buffer // len(records[0]["data"]) + 1

    # Buff the sum wf [pe] of non-saturated channels
    b_sumwf = np.zeros(len_buffer, dtype=np.float32)
    # Buff the records 'data' [ADC] in saturated channels
    b_pulse = np.zeros((n_channels, len_buffer), dtype=np.int16)
    # Buff the corresponding record index of saturated channels
    b_index = np.zeros((n_channels, max_nrecord), dtype=np.int64)

    # Main
    for ix, peak_i in enumerate(peak_list):
        # reset buffers
        b_sumwf[:] = 0
        b_pulse[:] = 0
        b_index[:] = -1

        p = peaks[peak_i]
        channel_saturated = p["saturated_channel"] > 0

        for record_i in range(record_ranges[ix][0], record_ranges[ix][1]):
            r = records[record_i]
            r_slice, b_slice = strax.overlap_indices(
                r["time"] // dt, r["length"], p["time"] // dt, p["length"] * p["dt"] // dt
            )

            ch = r["channel"]
            if channel_saturated[ch]:
                b_pulse[ch, slice(*b_slice)] += r["data"][slice(*r_slice)]
                b_index[ch, np.argmin(b_index[ch])] = record_i
            else:
                b_sumwf[slice(*b_slice)] += r["data"][slice(*r_slice)] * to_pe[ch]

        _peak_saturation_correction_inner(
            channel_saturated,
            records,
            p,
            to_pe,
            b_sumwf,
            b_pulse,
            b_index,
            reference_length,
            min_reference_length,
        )

        # Back track sum wf downsampling
        peaks[peak_i]["length"] = p["length"] * p["dt"] / dt
        peaks[peak_i]["dt"] = dt

    strax.sum_waveform(
        peaks,
        hitlets,
        records,
        rlinks,
        to_pe,
        n_top_channels=n_top_channels,
        store_data_top=store_data_top,
        store_data_start=store_data_start,
        select_peaks_indices=peak_list,
    )
    return peak_list


@numba.jit(nopython=True, nogil=True, cache=True)
def _peak_saturation_correction_inner(
    channel_saturated,
    records,
    p,
    to_pe,
    b_sumwf,
    b_pulse,
    b_index,
    reference_length=100,
    min_reference_length=20,
):
    """Would add a third level loop in peak_saturation_correction Which is not ideal for numba, thus
    this function is written.

    :param channel_saturated: (bool, n_channels)
    :param p: One peak/peaklet
    :param to_pe: adc to PE conversion (length should equal number of PMTs)
    :param b_sumwf b_pulse b_index: Filled buffers

    """
    dt = records["dt"][0]
    n_channels = len(channel_saturated)

    for ch in range(n_channels):
        if not channel_saturated[ch]:
            continue
        b = b_pulse[ch]

        # Define the reference region as reference_length before the first saturation point
        # unless there are not enough samples
        bl = np.inf
        for record_i in b_index[ch]:
            if record_i == -1:
                break
            bl = min(bl, records["baseline"][record_i])

        s0 = np.argmax(b >= np.int16(bl))
        ref = slice(max(0, s0 - reference_length), s0)

        if (b[ref] * to_pe[ch] > 1).sum() < min_reference_length:
            # the pulse is saturated, but there are not enough reference samples to get a good ratio
            # This actually distinguished between S1 and S2 and will only correct S2 signals
            continue
        if (b_sumwf[ref] > 1).sum() < min_reference_length:
            # the same condition applies to the waveform model
            continue
        if np.sum(b[ref]) * to_pe[ch] / np.sum(b_sumwf[ref]) > 1:
            # The pulse is saturated,
            # but insufficient information is available in the other channels
            # to reliably reconstruct it
            continue

        scale = np.sum(b[ref]) / np.sum(b_sumwf[ref])

        # Loop over the record indices of the saturated channel (saved in b_index buffer)
        for record_i in b_index[ch]:
            if record_i == -1:
                break
            r = records[record_i]
            r_slice, b_slice = strax.overlap_indices(
                r["time"] // dt, r["length"], p["time"] // dt + s0, p["length"] * p["dt"] // dt - s0
            )

            if r_slice[1] == r_slice[0]:  # This record proceeds saturation
                continue
            b_slice = b_slice[0] + s0, b_slice[1] + s0

            # First is finding the highest point in the desaturated record
            # because we need to bit shift the whole record if it exceeds int16 range
            apax = scale * max(b_sumwf[slice(*b_slice)])

            if np.int32(apax) >= 2**15:  # int16(2**15) is -2**15
                bshift = int(np.floor(np.log2(apax) - 14))

                tmp = r["data"].astype(np.int32)
                tmp[slice(*r_slice)] = b_sumwf[slice(*b_slice)] * scale

                r["area"] = np.sum(tmp)  # Auto covert to int64
                r["data"][:] = np.right_shift(tmp, bshift)
                r["amplitude_bit_shift"] += bshift
            else:
                r["data"][slice(*r_slice)] = b_sumwf[slice(*b_slice)] * scale
                r["area"] = np.sum(r["data"])


@numba.jit(nopython=True, nogil=True, cache=True)
def get_tight_coin(hit_max_times, hit_channel, peak_max_times, left, right, channels=(0, 493)):
    """Calculates the tight coincidence based on PMT channels.

    Defined by number of hits within a specified time range of the
    the peak's maximum amplitude.
    Imitates tight_coincidence variable in pax:
    github.com/XENON1T/pax/blob/master/pax/plugins/peak_processing/BasicProperties.py

    :param hit_max_times: Time of the hit amplitude in ns.
    :param hit_channel: PMT channels of the hits
    :param peak_max_times: Time of the peaks maximum in ns.
    :param left: Left boundary in which we search for the tight
        coincidence in ns.
    :param right: Right boundary in which we search for the tight
        coincidence in ns.
    :param channel_range: (min/max) channel for the corresponding detector.

    :return: n_coin_channel of length peaks containing the
        tight coincidence.

    """
    left_hit_i = 0
    n_coin_channel = np.zeros(len(peak_max_times), dtype=np.int16)
    start_ch, end_ch = channels
    channels_seen = np.zeros(end_ch - start_ch + 1, dtype=np.bool_)

    # loop over peaks
    for p_i, p_t in enumerate(peak_max_times):
        channels_seen[:] = 0
        # loop over hits starting from the last one we left at
        for left_hit_i in range(left_hit_i, len(hit_max_times)):
            # if the hit is in the window, its a tight coin
            d = hit_max_times[left_hit_i] - p_t
            if (-left <= d) & (d <= right):
                channels_seen[hit_channel[left_hit_i] - start_ch] = 1

            # stop the loop when we know we're outside the range
            if d > right:
                n_coin_channel[p_i] = np.sum(channels_seen)
                break

        # Add channel information in case there are no hits beyond
        # the last peak:
        n_coin_channel[p_i] = np.sum(channels_seen)

    return n_coin_channel



import numpy as np
import strax
import straxen
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO
from straxen.plugins.peaklets.peaklet_positions_cnf import PeakletPositionsCNF


export, __all__ = strax.exporter()


@export
class EventPositionContour(strax.Plugin):
    """A strax plugin that computes event position contours and applies field distortion corrections
    using the conditional normalising flow model. For information on the model, see note_.

    .. _note: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:juehang:flow_posrec_proposal_sr2 # noqa: E501

    This plugin calculates position contours for S2 and alternative S2 signals, and applies
    field distortion corrections to improve the accuracy of position reconstruction.

    Attributes:
        __version__ (str): Version of the plugin.
        depends_on (tuple): Input data types required for this plugin.
        provides (str): Output data type provided by this plugin.
        compressor (str): Compression algorithm used for data storage.
        data_kind (str): Kind of data this plugin processes.
        loop_over (str): Data structure to iterate over during computation.

    Configuration options:
        n_poly (int): Number of points in the uncertainty contour.
        electron_drift_velocity (float): Electron drift velocity in cm/ns.
        fdc_map (str): Path to the 3D field distortion correction map.
        use_fdc_for_contour (bool): Whether to use full FDC for
        position reconstruction uncertainty contours.

    """

    __version__ = "0.1.0"

    depends_on = ("event_basics", "event_positions", "peak_positions_cnf", "peak_basics")

    provides = "event_position_contour"
    compressor = "zstd"
    data_kind = "events"
    loop_over = "events"

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    n_poly = straxen.URLConfig(
        default=16,
        infer_type=False,
        help="Size of uncertainty contour",
    )

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    fdc_map = straxen.URLConfig(
        default="xedocs://fdc_maps"
        "?algorithm=plugin.default_reconstruction_algorithm&run_id=plugin.run_id"
        "&attr=map&scale_coordinates=plugin.coordinate_scale&version=ONLINE",
        infer_type=False,
        help="3D field distortion correction map path",
    )

    use_fdc_for_contour = straxen.URLConfig(
        infer_type=False,
        help=(
            "Whether to use full FDC for position reconstruction uncertainty contours. "
            "Uses the FDC of the central point if false."
        ),
        default=False,
    )

    def infer_dtype(self):
        """Infer the data type for the output array.

        Returns:
            list: A list of tuples defining the structure of the output array.

        """
        # Define information lines for different signal types
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        dtype = []

        # Add fields for position contours
        ptypes = ["s2", "alt_s2"]
        for type_ in ptypes:
            dtype += [
                (
                    (
                        f"Position uncertainty contour for {infoline[type_]}",
                        f"{type_}_position_contour_cnf",
                    ),
                    np.float32,
                    (self.n_poly + 1, 2),
                )
            ]

        # Add fields for flow position contour and corrected positions
        dtype += [
            (
                (
                    "Position uncertainty contour, field-distortion corrected (cm)",
                    "position_contour_cnf_fdc",
                ),
                np.float32,
                (self.n_poly + 1, 2),
            ),
            (
                ("Reconstructed cnf S2 X position, field-distortion corrected (cm)", "x_cnf_fdc"),
                np.float32,
            ),
            (
                ("Reconstructed cnf S2 Y position, field-distortion corrected (cm)", "y_cnf_fdc"),
                np.float32,
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        """Set up the plugin by initializing coordinate scales and loading the FDC map."""
        self.coordinate_scales = [1.0, 1.0, -self.electron_drift_velocity]
        self.map = self.fdc_map

    def compute(self, events, peaks):
        """Compute event position contours and apply field distortion corrections.

        Args:
            events (np.ndarray): Input events array.
            peaks (np.ndarray): Input peaks array.

        Returns:
            np.ndarray: Array containing computed position contours and corrected positions.

        """
        # Initialize the result array
        result = np.zeros(len(events), dtype=self.dtype)
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)

        # Initialize contour fields with NaN values
        for type_ in ["s2", "alt_s2"]:
            result[f"{type_}_position_contour_cnf"] *= np.nan
        result["position_contour_cnf_fdc"] *= np.nan

        # Split peaks by containment in events
        split_peaks = strax.split_by_containment(peaks, events)

        # Process each event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ["s2", "alt_s2"]:
                type_index = event[f"{type_}_index"]
                if type_index != -1:
                    # Store naive flow position contour
                    result[f"{type_}_position_contour_cnf"][event_i] = sp["position_contour_cnf"][
                        type_index
                    ]

                if type_ == "s2":
                    if self.use_fdc_for_contour:
                        # Apply full field distortion correction to contour
                        contour_plus_xy = np.concatenate(
                            [
                                sp["position_contour_cnf"][type_index],
                                np.array([[sp["x_cnf"][type_index], sp["y_cnf"][type_index]]]),
                            ],
                            axis=0,
                        )
                        contour_with_z = np.concatenate(
                            [
                                contour_plus_xy,
                                np.repeat(event["z_dv_corr"], self.n_poly + 2)[:, np.newaxis],
                            ],
                            axis=1,
                        )
                        delta_r = self.map(contour_with_z)
                        scale = delta_r / np.linalg.norm(contour_with_z[:, :2], axis=1) + 1
                        scaled_2d_contour = contour_with_z[:, :2] * scale[:, np.newaxis]

                        # Store corrected contour and positions
                        result["position_contour_cnf_fdc"][event_i] = scaled_2d_contour[:-1]
                        result["x_cnf_fdc"][event_i] = scaled_2d_contour[-1, 0]
                        result["y_cnf_fdc"][event_i] = scaled_2d_contour[-1, 1]
                    else:
                        # Apply simple scaling based on field distortion correction
                        scale = event["r_field_distortion_correction"] / event["r_naive"] + 1
                        result["position_contour_cnf_fdc"][event_i] = (
                            sp["position_contour_cnf"][type_index] * scale
                        )
                        result["x_cnf_fdc"][event_i] = sp["x_cnf"][type_index] * scale
                        result["y_cnf_fdc"][event_i] = sp["y_cnf"][type_index] * scale

        return result


@export
class EventPositionUncertainty(strax.Plugin):
    """Plugin to calculate position uncertainties for events using the conditional normalising flow
    model. For information on the model, see note_.

    .. _note: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:juehang:flow_posrec_proposal_sr2 # noqa: E501

    This plugin computes uncertainties in radial (r) and angular (theta) positions
    for both main and alternative S2 signals. It uses the position contours
    calculated by a flow-based algorithm to estimate these uncertainties.

    The plugin provides:
    - Naive flow position uncertainties for main and alternative S2 signals
    - Corrected flow position uncertainties accounting for field distortions

    Depends on:
    - 'events': Basic event information
    - 'event_info': Additional event information
    - 'event_position_contour': Position contours for events

    Provides:
    - 'event_position_uncertainty': Calculated position uncertainties

    """

    __version__ = "0.0.2"

    depends_on = ("event_info", "event_positions", "event_position_contour")
    provides = "event_position_uncertainty"

    def infer_dtype(self):
        # Define the data types for the output
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        dtype = []
        # Add fields for position uncertainties
        ptypes = ["s2", "alt_s2"]
        for type_ in ptypes:
            dtype += [
                (
                    (
                        f"Naive flow position uncertainty in r for {infoline[type_]}",
                        f"{type_}_r_position_uncertainty",
                    ),
                    np.float32,
                ),
                (
                    (
                        f"Naive flow position uncertainty in theta for {infoline[type_]}",
                        f"{type_}_theta_position_uncertainty",
                    ),
                    np.float32,
                ),
            ]

        # Add fields for corrected position uncertainties
        dtype += [
            (
                ("Position uncertainty in r (cm)", "r_position_uncertainty"),
                np.float32,
            ),
            (
                ("Position uncertainty in theta (rad)", "theta_position_uncertainty"),
                np.float32,
            ),
        ]
        # Add time fields
        dtype += strax.time_fields
        return dtype

    def compute(self, events):
        # Initialize the result array
        result = np.zeros(len(events), dtype=self.dtype)
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)

        # Initialize uncertainty fields with NaN
        for type_ in ["s2", "alt_s2"]:
            result[f"{type_}_r_position_uncertainty"] *= np.nan
            result[f"{type_}_theta_position_uncertainty"] *= np.nan
        result["r_position_uncertainty"] *= np.nan
        result["theta_position_uncertainty"] *= np.nan

        # Calculate uncertainties for main and alternative S2 signals
        for type_ in ["s2", "alt_s2"]:
            # Calculate radial uncertainties
            r_array = np.linalg.norm(events[f"{type_}_position_contour_cnf"], axis=2)
            r_min = np.min(r_array, axis=1)
            r_max = np.max(r_array, axis=1)

            # Calculate angular uncertainties
            theta_array = np.arctan2(
                events[f"{type_}_position_contour_cnf"][..., 1],
                events[f"{type_}_position_contour_cnf"][..., 0],
            )

            avg_theta = np.arctan2(events[f"{type_}_y_cnf"], events[f"{type_}_x_cnf"])

            theta_diff = PeakletPositionsCNF.calculate_theta_diff(theta_array, avg_theta)

            # Store uncertainties
            result[f"{type_}_r_position_uncertainty"] = (r_max - r_min) / 2
            result[f"{type_}_theta_position_uncertainty"] = np.abs(theta_diff) / 2

        # Apply field distortion correction to uncertainties
        scale = events["r_field_distortion_correction"] / events["r_naive"] + 1
        result["r_position_uncertainty"] = result["s2_r_position_uncertainty"] * scale
        result["theta_position_uncertainty"] = result["s2_theta_position_uncertainty"] * scale

        return result



import strax
import straxen
from straxen.analyses.holoviews_waveform_display import seconds_from

import panel as pn
from bokeh.models import HoverTool

import numpy as np
import pandas as pd

export, __all__ = strax.exporter()


@export
class nVETOEventDisplay:
    def __init__(
        self,
        events=None,
        hitlets=None,
        run_id=0,
        channel_range=(2000, 2119),
        pmt_map="nveto_pmt_position.csv",
        plot_extension="bokeh",
    ):
        """Class to plot an interactive nveto display.

        :param events: Events which should be plot. Can also be none in case the hitlet matrix
            and/or pattern map should be plotted separately.
        :param hitlets: Same as events, but hitlets_nv.
        :param run_id: Run_id which should be displayed in the title.
        :param channel_range: Channel range of the detector.
        :param pmt_map: PMT map which is loaded via straxen.get_resource. The map has to contain the
            channel number, and xyz coordinates.
        :param plot_extension: Extension which should be used for rendering can be either bokeh or
            matpltolib. Default is bokeh to support dynamic plots.

        """
        self.import_holoviews()
        self.hv.extension(plot_extension)
        self.df_event_time = None
        self.df_event_properties = None
        self.hitlets = hitlets
        self.channel_range = channel_range
        self.run_id = run_id

        # Load PMT data:
        if isinstance(pmt_map, str):
            self.pmt_positions = straxen.get_resource(pmt_map, fmt="csv")
        elif isinstance(pmt_map, np.ndarray):
            self.pmt_positions = pmt_map
        else:
            raise ValueError(
                "pmt_map not understood, has either to be "
                f'a string or a numpy array, got "{pmt_map}".'
            )

        if events is not None:
            self.event_df = straxen.convert_array_to_df(events)
        else:
            self.event_df = None

        if events is not None and hitlets is not None:
            self.hitlets_per_event = strax.split_by_containment(hitlets, events)

    def import_holoviews(self):
        import holoviews as hv

        self.hv = hv

    def plot_event_display(self):
        """Creates an interactive event display for the neutron veto.

        :return: panel.Column hosting the plots and panels.

        """

        # First we have to define the python callbacks:
        def matrix_callback(value):
            """Callback for the dynamic hitlet matrix.

            Changes polygons when a new event is selected.

            """
            self.hitlet_points = self.hitlets_to_hv_points(
                self.hitlets_per_event[value], t_ref=self.event_df.loc[value, "time"]
            )

            # Create the hitlet matrix and time stream:
            self.hitlet_matrix = self.plot_hitlet_matrix(
                hitlets=None, _hitlet_points=self.hitlet_points
            )
            return self.hitlet_matrix

        def pattern_callback(value, x_range):
            """Call back for the dynamic PMT pattern map.

            Depends on the selcted event as well as the selected x_range in the hitlet_matrix.

            """
            # Get hitlet points and select only points within x_range:
            hit = self.hitlet_points.data
            if not x_range:
                time = hit["time"].values
                length = hit["length"].values
                dt = hit["dt"].values
                x_range = [min(time), max(time + (length * dt))]

            m = (hit["time"] >= x_range[0]) & (hit["time"] < x_range[1])

            hitlets_in_time = hit[m]
            new_points = self.hv.Points(hitlets_in_time)

            # Plot pmt pattern:
            pmts = self.plot_nveto(
                hitlets=None, _hitlet_points=new_points, pmt_size=8, pmt_distance=0.5
            )
            angle = self._plot_reconstructed_position(value)
            return angle * pmts

        self._make_sliders_and_tables(self.event_df)
        index = self.evt_sel_slid.value
        self.hitlet_points = self.hitlets_to_hv_points(
            self.hitlets_per_event[index], t_ref=self.event_df.loc[index, "time"]
        )

        dmap_hitlet_matrix = self.hv.DynamicMap(
            matrix_callback, streams=[self.evt_sel_slid.param.value]
        ).opts(framewise=True)

        time_stream = self.hv.streams.RangeX(source=dmap_hitlet_matrix)

        dmap_pmts = self.hv.DynamicMap(
            pattern_callback, streams=[self.evt_sel_slid.param.value, time_stream]
        )

        slider_column = pn.Column(
            self.evt_sel_slid, self.evt_sel_slid.controls(["value"]), self.time_table
        )

        event_display = pn.Column(
            self.title_panel,
            pn.Row(slider_column, dmap_pmts, width_policy="max"),
            dmap_hitlet_matrix,
            self.prop_table,
            self.pos_table,
            width_policy="max",
        )
        return event_display

    def plot_hitlet_matrix(self, hitlets, _hitlet_points=None):
        """Function which plots the hitlet matrix for the specified hitlets. The hitlet matrix is
        something equivalent to the record matrix for the TPC.

        :param hitlets: Hitlets to be plotted if called directly.
        :param _hitlet_points: holoviews.Points created by the event display. Only internal use.
        :return: hv.Polygons plot.

        """
        if not _hitlet_points:
            _hitlet_points = self.hitlets_to_hv_points(
                hitlets,
            )

        hitlet_matrix = self._plot_base_matrix(_hitlet_points).opts(
            title="Hitlet Matrix",
            xlabel="Time [ns]",
            ylabel="PMT channel",
            ylim=(1995, 2125),
            color="area",
            clabel="Area [pe]",
            cmap="viridis",
            colorbar=True,
        )
        return hitlet_matrix

    def _plot_base_matrix(self, hv_point_data):
        """Base function to plot record or hitlet matrix."""
        matrix_plot = self.hv.Segments(
            hv_point_data.data, kdims=["time", "channel", "endtime", "channel"]
        ).opts(
            tools=["hover"],
            aspect=4,
            responsive="width",
        )
        return matrix_plot

    def plot_nveto(
        self,
        hitlets,
        pmt_size=8,
        pmt_distance=0.5,
        _hitlet_points=None,
    ):
        """Plots the nveto pmt pattern map for the specified hitlets. Expects hitlets to be sorted
        in time.

        :param hitlets: Hitlets to be plotted if called directly.
        :param pmt_size: Base size of a PMT for 1 pe.
        :param pmt_distance: Scaling parameter for the z -> xy projection.
        :param _hitlet_points: holoviews.Points created by the event display. Only internal use.
        :return: stacked hv.Points plot.

        """
        if not _hitlet_points:
            _hitlet_points = self.hitlets_to_hv_points(
                hitlets,
            )

        pmts = self._plot_nveto(_hitlet_points.data, pmt_size=pmt_size, pmt_distance=pmt_distance)

        # Plot channels which have not seen anything:
        pmts_data = pmts.data
        m = pmts_data.loc[:, "area"].values == 0
        blanko_pmts = np.zeros(np.sum(m), dtype=strax.hitlet_dtype())
        blanko_pmts["area"] = 1
        blanko_pmts["channel"] = pmts_data.loc[m, "channel"]
        blanko = self._plot_nveto(blanko_pmts, pmt_size=pmt_size, pmt_distance=pmt_distance)

        blanko = blanko.opts(
            fill_color="white", tools=[HoverTool(tooltips=[("PMT", "@channel")])], color=None
        )

        detector_layout = (
            plot_tpc_circle(straxen.cryostat_outer_radius)
            * plot_diffuser_balls_nv()
            * plot_nveto_reflector()
        )

        return detector_layout * blanko * pmts

    def _plot_nveto(self, pmts, pmt_size, pmt_distance):
        if isinstance(pmts, pd.DataFrame):
            pmts = pmts.to_records(index=False)
        hv = self.hv
        # Get hitlet_points data and extract PMT data
        pmt_data = self._create_nveto_pmts(pmts, pmt_size, pmt_distance)

        pmts = hv.Points(
            data=pd.DataFrame(pmt_data),
            kdims=[hv.Dimension("pmt_x", label="x [cm]"), hv.Dimension("pmt_y", label="y [cm]")],
            vdims=["time", "size", "area", "channel"],
        ).opts(
            size="size",
            tools=[self._create_pmt_pattern_hover()],
            color="time",
            colorbar=True,
            cmap="viridis",
            clabel="Time [ns]",
            line_color="black",
            alpha=0.8,
            width=400,
            title="nVETO Top View",
        )

        return pmts

    def _create_nveto_pmts(
        self,
        hitlets,
        pmt_size,
        pmt_distance,
        max_area_scale=10,
    ):
        """Function which creates data for the nVTEO PMT points of the pattern plot."""
        # Get PMT xy-position based on projection function:
        pmt_data = self._convert_channel_to_xy(pmt_distance=pmt_distance)

        for h in hitlets:
            ch = h["channel"] - self.channel_range[0]
            pmt_data[ch]["area"] += h["area"]

            if not pmt_data[ch]["time"]:
                # Get first arrival time for channel:
                pmt_data[ch]["time"] = h["time"]

        pmt_data["size"] = pmt_data["area"] * pmt_size
        pmt_data["size"] = np.clip(pmt_data["size"], 0, max_area_scale * pmt_size)

        return pmt_data

    def _convert_channel_to_xy(self, pmt_distance=0.5):
        """Projects PMT channel from x/y/z onto xy."""
        pmt_data = np.zeros(
            120,
            dtype=[
                ("pmt_x", np.float64),
                ("pmt_y", np.float64),
                ("channel", np.int32),
                ("column", np.int32),
                ("area", np.float64),
                ("size", np.float64),
                ("time", np.float64),
            ],
        )
        pmt_data["pmt_x"] = self.pmt_positions["x"]
        pmt_data["pmt_y"] = self.pmt_positions["y"]
        pmt_data["channel"] = self.pmt_positions["channel"]
        pmt_data["column"] = np.repeat(np.arange(20), 6)

        z = self.pmt_positions["z"]
        z_min = self.pmt_positions["z"].min()
        z_distance = self.pmt_positions["z"].max() - z_min
        scale = (z - z_min) / z_distance * pmt_distance

        # Only correct x or y depending on column position:
        m_x = np.isin(pmt_data["column"], (4, 5, 6, 7, 8, 14, 15, 16, 17, 18))
        m_y = np.isin(pmt_data["column"], (19, 0, 1, 2, 3, 9, 10, 11, 12, 13))

        pmt_data["pmt_x"][m_x] = pmt_data["pmt_x"][m_x] * (1 + scale[m_x])
        pmt_data["pmt_y"][m_y] = pmt_data["pmt_y"][m_y] * (1 + scale[m_y])

        return pmt_data

    @staticmethod
    def _create_pmt_pattern_hover():
        tooltips = [
            ("rel. time [ns]", "@time"),
            ("integrated area [pe]", "@area"),
            ("PMT", "@channel"),
        ]
        return HoverTool(tooltips=tooltips)

    @staticmethod
    def hitlets_to_hv_points(
        hitlets,
        t_ref=None,
    ):
        """Function which converts hitlets into hv.Points used in the different plots.

        Computes hitlet times as relative times with respect to the first hitlet if t_ref is not
        set.

        """
        import holoviews as hv

        if not len(hitlets):
            raise ValueError("Expected at least a single hitlet.")

        if isinstance(hitlets, np.ndarray):
            hitlets = pd.DataFrame(hitlets)

        # Set relative times:
        if t_ref is None:
            t_ref = min(hitlets["time"])

        time = seconds_from(hitlets["time"], t_ref, unit_conversion=1)
        hitlets["time"] = time
        hitlets["endtime"] = strax.endtime(hitlets.to_records())

        hitlet_points = hv.Points(hitlets)

        return hitlet_points

    def _plot_reconstructed_position(self, index):
        """Function which plots the nVETO event position according to its azimuthal angle.

        :param index: Which event to plot.

        """

        x = (0, np.real(np.exp(self.event_df.loc[index, "angle"] * 1j)) * 400)
        y = (0, np.imag(np.exp(self.event_df.loc[index, "angle"] * 1j)) * 400)
        angle = self.hv.Curve((x, y)).opts(
            color="orange", line_dash="dashed", xlim=(-350, 350), ylim=(-350, 350)
        )
        return angle

    def _make_sliders_and_tables(self, df):
        """Function which creates interactive sliders and tables for the neutron-veto event
        display."""
        if not len(df):
            raise ValueError("DataFrame must be at least one entry long.")

        self.evt_sel_slid = pn.widgets.IntSlider(value=0, start=0, end=len(df))
        self._make_tables(df)

        # Define callbacks for tables:
        self.evt_sel_slid.param.watch(
            lambda event: table_callback(self.time_table, self.df_event_time, event, True), "value"
        )
        self.evt_sel_slid.param.watch(
            lambda event: table_callback(self.prop_table, self.df_event_properties, event), "value"
        )
        self.evt_sel_slid.param.watch(
            lambda event: table_callback(self.pos_table, self.df_event_position, event), "value"
        )

        # Now make title and also define callback:
        title = self._make_title(self.evt_sel_slid.value)
        self.title_panel = pn.panel(title, sizing_mode="scale_width")

        def title_callback(event):
            self.title_panel.object = self._make_title(event.new)

        self.evt_sel_slid.param.watch(title_callback, "value")

    def _make_tables(self, df):
        """Divides event data into a time, properties and position table."""
        # Time table:
        time_keys = ["time", "endtime", "event_number_nv"]
        self.df_event_time = df.loc[:, time_keys]

        # Properties tables:
        pos_keys = [
            "angle",
            "pos_x",
            "pos_x_spread",
            "pos_y",
            "pos_y_spread",
            "pos_z",
            "pos_z_spread",
        ]
        self.df_event_position = df.loc[:, pos_keys]

        keys = df.columns.values
        keys = [k for k in keys if k not in time_keys + pos_keys]
        self.df_event_properties = df.loc[:, keys]

        # Table panels:
        index = self.evt_sel_slid.value
        self.time_table = pn.panel(
            self.df_event_time.loc[index],
        )
        self.pos_table = pn.panel(
            self.df_event_position.loc[index:index, :], sizing_mode="scale_width"
        )

        self.prop_table = pn.panel(
            self.df_event_properties.loc[index:index, :], sizing_mode="scale_width"
        )

    def _make_title(self, ind):
        """Function which creates title test in markdown format."""
        start = self.df_event_time.loc[ind, "time"]
        date = np.datetime_as_string(start.astype("<M8[ns]"), unit="s")
        start_ns = start - (start // straxen.units.s) * straxen.units.s
        end = self.df_event_time.loc[ind, "endtime"]
        end_ns = end - start + start_ns
        return "".join(
            (
                f"##Event {ind} from run {self.run_id}\n",
                f"##Recorded at ({date[:10]} {date[10:]}) UTC ",
                f"{start_ns} ns - {end_ns} ns",
            )
        )


def plot_tpc_circle(radius):
    """Plots TPC as a black circle.

    :param radius: Radius in cm.

    """
    import holoviews as hv

    x = radius * np.cos(np.arange(-np.pi, np.pi + 0.1, 0.01))
    y = radius * np.sin(np.arange(-np.pi, np.pi + 0.1, 0.01))
    return hv.Curve((x, y)).opts(color="k")


def plot_diffuser_balls_nv():
    """Computes position of nveto diffuser balls.

    :return: hv.Points with hover tool.

    """
    import holoviews as hv

    cryostat_r = straxen.cryostat_outer_radius
    depth_stiffening_ring = 8
    r = cryostat_r + depth_stiffening_ring
    angles = np.array([-8, 90, -90, 180])
    x = r * np.cos(angles / 180 * np.pi)
    y = r * np.sin(angles / 180 * np.pi)
    db_ids = (18, 17, 11, 15)
    data = pd.DataFrame(np.array([x, y, db_ids]).T, columns=("x", "y", "id"))
    return hv.Points(data=data).opts(size=8, tools=["hover"])


def plot_nveto_reflector():
    """Function which plots the lateral reflector panels of the neutron veto.

    Coordinates are based on the MC coordinates of the octagon
    model. The coordinates can be found in this note:
    id=xenon:xenonnt:mc:notes:nveto-geometry#lateral_reflectors

    """
    import holoviews as hv

    xy_center_angle = np.array(
        [
            (0, -1875.25, -90),
            (1442.28, -1442.28, -45),
            (1875.25, 0, 0),
            (1442.28, 1442.28, 45),
            (0, 1875.25, 90),
            (-1442.28, 1442.28, 135),
            (-1875.25, 0, 180),
            (-1441.75, -1442.28, -135),
            (0, -1875.25, -90),
        ],
        dtype=[
            ("x", np.float64),
            ("y", np.float64),
            ("phi", np.float64),
        ],
    )
    xy_pos = _compute_lateral_reflector_xy_edges(
        xy_center_angle, long_side_length=2018, short_side_length=1224
    )
    xy_pos = pd.DataFrame(xy_pos, columns=("x", "y"))

    return hv.Curve(data=xy_pos).opts(color="k")


def _compute_lateral_reflector_xy_edges(
    xy_center_angle, long_side_length=2018, short_side_length=1224
):
    """Function which computes the position of the lateral reflector panels. Input in mm return in
    cm.

    :param xy_center_angle: Center xy coordinate and angle of each panel.
    :param long_side_length: Full length of the long panels in mm.
    :param short_side_length: Same

    """
    res = np.zeros(len(xy_center_angle), dtype=[("x", np.float64), ("y", np.float64)])
    for ind, xy in enumerate(xy_center_angle):
        if ind % 2:
            length = short_side_length / 2
        else:
            length = long_side_length / 2
        xc, yc, phi = xy[["x", "y", "phi"]]
        x = xc - length * np.sin(phi * np.pi / 180)
        y = yc + length * np.cos(phi * np.pi / 180)
        res[ind]["x"] = x / 10
        res[ind]["y"] = y / 10

    return res


def table_callback(table, data, event, column=False):
    """Callback template for tables used together with a pn.widget.IntSlider.

    :param table: pn.Panel object of a pd.DataFrame
    :param data: pd.DataFrame storing the data.
    :param event: Slider event returned by param.watch
    :param column: Boolean if true uses horizontal columns to present data.

    """
    if column:
        table.object = data.loc[event.new]
    else:
        table.object = data.loc[event.new : event.new]



import strax
from straxen.plugins.peaklets.peaklet_positions_mlp import PeakletPositionsMLP

export, __all__ = strax.exporter()


@export
class MergedS2PositionsMLP(PeakletPositionsMLP):

    __version__ = "0.0.0"
    child_plugin = True
    algorithm = "mlp"
    depends_on = "merged_s2s"
    provides = "merged_s2_positions_mlp"

    def compute(self, merged_s2s):
        return super().compute(merged_s2s)



import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import strax
import straxen

export, __all__ = strax.exporter()


@export
def plot_pmts(
    c,
    label="",
    figsize=None,
    show_tpc=True,
    extend="neither",
    vmin=None,
    vmax=None,
    **kwargs,
):
    """Plot the PMT arrays side-by-side, coloring the PMTS with c.

    :param c: Array of colors to use. Must have len() n_tpc_pmts
    :param label: Label for the color bar
    :param figsize: Figure size to use.
    :param extend: same as plt.colorbar(extend=...)
    :param vmin: Minimum of color scale
    :param vmax: maximum of color scale
    :param show_axis_labels: if True it will show x and y labels
    Other arguments are passed to plot_on_single_pmt_array.

    """
    if vmin is None:
        vmin = np.nanmin(c)
    if vmax is None:
        vmax = np.nanmax(c)
    if vmin == vmax:
        # Single-valued array passed
        vmax += 1
    if figsize is None:
        figsize = (13.25, 5.75)

    f, axes = plt.subplots(1, 2, figsize=figsize)
    plot_result = None
    for array_i, array_name in enumerate(["top", "bottom"]):
        ax = axes[array_i]
        plt.sca(ax)
        plt.title(array_name.capitalize())

        plot_result = plot_on_single_pmt_array(
            c,
            array_name=array_name,
            show_tpc=show_tpc,
            vmin=vmin,
            vmax=vmax,
            **kwargs,
        )

    axes[0].set_xlabel("x [cm]")
    axes[0].xaxis.set_label_coords(1.035, -0.075)
    axes[0].set_ylabel("y [cm]")

    axes[1].yaxis.tick_right()
    axes[1].yaxis.set_label_position("right")

    plt.tight_layout()
    plt.subplots_adjust(wspace=0)
    plt.colorbar(mappable=plot_result, ax=axes, extend=extend, label=label)


@export
def plot_on_single_pmt_array(
    c,
    array_name="top",
    r=straxen.tpc_r * 1.03,
    pmt_label_size=8,
    pmt_label_color="white",
    show_tpc=True,
    log_scale=False,
    vmin=None,
    vmax=None,
    dead_pmts=None,
    dead_pmt_color="gray",
    **kwargs,
):
    """Plot one of the PMT arrays and color it by c.

    :param c: Array of colors to use. Must be len() of the number of TPC PMTs
    :param label: Label for the color bar
    :param pmt_label_size: Fontsize for the PMT number labels.
    Set to 0 to disable.
    :param pmt_label_color: Text color of the PMT number labels.
    :param log_scale: If True, use a logarithmic color scale
    :param extend: same as plt.colorbar(extend=...)
    :param vmin: Minimum of color scale
    :param vmax: maximum of color scale
    Other arguments are passed to plt.scatter.

    """
    if vmin is None:
        vmin = c.min()
    if vmax is None:
        vmax = c.max()

    pmt_positions = straxen.pmt_positions().to_records()

    ax = plt.gca()
    ax.set_aspect("equal")
    plt.xlim(-r, r)
    plt.ylim(-r, r)

    mask = pmt_positions["array"] == array_name
    pos = pmt_positions[mask]

    kwargs.setdefault("s", 280)
    if log_scale:
        kwargs.setdefault("norm", matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax))
    else:
        kwargs.setdefault("vmin", vmin)
        kwargs.setdefault("vmax", vmax)
    result = plt.scatter(pos["x"], pos["y"], c=c[mask], **kwargs)

    if show_tpc:
        ax.set_facecolor("lightgrey")
        ax.add_artist(
            plt.Circle((0, 0), straxen.tpc_r, edgecolor="k", facecolor="w", zorder=-5, linewidth=1)
        )
    else:
        ax.set_axis_off()
    if dead_pmts is not None:
        _dead_mask = [pi in dead_pmts for pi in pos["i"]]
        plt.scatter(pos[_dead_mask]["x"], pos[_dead_mask]["y"], c=dead_pmt_color, **kwargs)

    if pmt_label_size:
        for p in pos:
            plt.text(
                p["x"],
                p["y"],
                str(p["i"]),
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=pmt_label_size,
                color=pmt_label_color,
            )
    return result


@export
def log_y(a=None, b=None, scalar_ticks=True, tick_at=None):
    """Make the y axis use a log scale from a to b."""
    plt.yscale("log")
    if a is not None:
        if b is None:
            a, b = a[0], a[-1]
        ax = plt.gca()
        plt.ylim(a, b)
        if scalar_ticks:
            ax.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter("%g"))
            ax.set_yticks(logticks(a, b, tick_at))


@export
def log_x(a=None, b=None, scalar_ticks=True, tick_at=None):
    """Make the x axis use a log scale from a to b."""
    plt.xscale("log")
    if a is not None:
        if b is None:
            a, b = a[0], a[-1]
        plt.xlim(a, b)
        ax = plt.gca()
        if scalar_ticks:
            ax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter("%g"))
            ax.set_xticks(logticks(a, b, tick_at))


def logticks(tmin, tmax=None, tick_at=None):
    if tick_at is None:
        tick_at = (1, 2, 5, 10)
    a, b = np.log10([tmin, tmax])
    a = np.floor(a)
    b = np.ceil(b)
    ticks = strax.stable_sort(
        np.unique(np.outer(np.array(tick_at), 10.0 ** np.arange(a, b)).ravel())
    )
    ticks = ticks[(tmin <= ticks) & (ticks <= tmax)]
    return ticks


@export
def draw_box(x, y, **kwargs):
    """Draw rectangle, given x-y boundary tuples."""
    plt.gca().add_patch(
        matplotlib.patches.Rectangle(
            (x[0], y[0]), x[1] - x[0], y[1] - y[0], facecolor="none", **kwargs
        )
    )


@export
def plot_single_pulse(records, run_id, pulse_i=""):
    """Function which plots a single pulse.

    :param records: Records which belong to the pulse.
    :param run_id: Id of the run.
    :param pulse_i: Index of the pulse to be plotted.
    :return: fig, axes objects.

    """
    pulse = _make_pulse(records)

    fig, axes = plt.subplots()
    sec, ns = _split_off_ns(records[0]["time"])
    date = np.datetime_as_string(sec.astype("<M8[ns]"), unit="s")
    plt.title(f"Pulse {pulse_i} from {run_id}\nRecorded at {date[:10]}, {date[10:]} UTC {ns} ns")

    plt.step(np.arange(len(pulse)), pulse, where="post", label=f'Ch: {records[0]["channel"]}')
    plt.legend()
    plt.xlabel(f'Sample [{records[0]["dt"]} ns]')
    plt.ylabel("Height [ADC counts]")
    plt.grid()
    return fig, axes


def _make_pulse(records):
    """Helper to make a pulse based on fragements."""
    pulse = np.zeros(records[0]["pulse_length"], dtype=np.float32)

    offset = 0
    for r in records:
        pulse[offset : offset + r["length"]] = r["data"][: r["length"]]
        offset += r["length"]

    return pulse


def _split_off_ns(time):
    """Mini helper to divide time into seconds and ns."""
    sec = (time // 10**9) * 10**9
    ns = time - sec
    return sec, ns



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EventTopBottomParams(strax.Plugin):
    """Pluging that computes timing characteristics of top and bottom waveforms based on waveforms
    stored at event level for main/alt S1/S2."""

    depends_on = ("event_basics", "event_waveform")
    provides = "event_top_bottom_params"
    __version__ = "0.0.1"

    def infer_dtype(self):
        # Populating data type information
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        ev_info_fields = self.deps["event_basics"].dtype.fields
        dtype = []
        # populating APC and waveform samples
        self.ptypes = ["s1", "s2", "alt_s1", "alt_s2"]
        self.arrs = ["top", "bot"]
        for type_ in self.ptypes:
            for arr_ in self.arrs:
                dtype += [
                    (
                        (
                            f"Central time for {infoline[type_]} for {arr_} PMTs [ns]",
                            f"{type_}_center_time_{arr_}",
                        ),
                        ev_info_fields[f"{type_}_center_time"][0],
                    )
                ]
                dtype += [
                    (
                        (
                            (
                                f"Time between 10% and 50% area quantiles for {infoline[type_]} for"
                                f" {arr_} PMTs [ns]"
                            ),
                            f"{type_}_rise_time_{arr_}",
                        ),
                        ev_info_fields[f"{type_}_rise_time"][0],
                    )
                ]
                dtype += [
                    (
                        (
                            (
                                f"Width (in ns) of the central 50% area of the peak for {arr_} PMTs"
                                f" of {infoline[type_]}"
                            ),
                            f"{type_}_range_50p_area_{arr_}",
                        ),
                        ev_info_fields[f"{type_}_range_50p_area"][0],
                    )
                ]
                dtype += [
                    (
                        (
                            (
                                f"Width (in ns) of the central 90% area of the peak for {arr_} PMTs"
                                f" of {infoline[type_]}"
                            ),
                            f"{type_}_range_90p_area_{arr_}",
                        ),
                        ev_info_fields[f"{type_}_range_90p_area"][0],
                    )
                ]
            dtype += [
                (
                    (
                        (
                            "Difference between center times of top and bottom arrays for"
                            f" {infoline[type_]} [ns]"
                        ),
                        f"{type_}_center_time_diff_top_bot",
                    ),
                    ev_info_fields[f"{type_}_center_time"][0],
                )
            ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events):
        result = np.zeros(events.shape, dtype=self.dtype)
        if not len(events):
            return result

        result["time"], result["endtime"] = events["time"], strax.endtime(events)
        peak_dtype = strax.peak_dtype(n_channels=straxen.n_tpc_pmts, store_data_top=False)
        for type_ in self.ptypes:
            for arr_ in self.arrs:
                # in order to reuse the same definitions as in other parts, we create "fake peaks"
                # based only on data from corresponding array
                fpeaks_ = np.zeros(events.shape[0], dtype=peak_dtype)
                if arr_ == "top":
                    fpeaks_["data"] = events[f"{type_}_data_top"]
                    fpeaks_["area"] = events[f"{type_}_area"] * events[f"{type_}_area_fraction_top"]
                elif arr_ == "bot":
                    fpeaks_["data"] = events[f"{type_}_data"] - events[f"{type_}_data_top"]
                    fpeaks_["area"] = events[f"{type_}_area"] * (
                        1.0 - events[f"{type_}_area_fraction_top"]
                    )
                elif arr_ == "tot":
                    # This one is ony
                    fpeaks_["data"] = events[f"{type_}_data"]
                    fpeaks_["area"] = events[f"{type_}_area"]
                else:
                    raise RuntimeError(f"Received unknown array type : " + arr_)
                fpeaks_["length"] = events[f"{type_}_length"]
                fpeaks_["dt"] = events[f"{type_}_dt"]
                # computing central times
                # note that here we ignore 1/2 sample length to be consistent with other definitions
                with np.errstate(divide="ignore", invalid="ignore"):
                    recalc_ctime = np.sum(
                        fpeaks_["data"] * (np.arange(0, fpeaks_["data"].shape[1])), axis=1
                    )
                    recalc_ctime /= fpeaks_["area"]
                    recalc_ctime *= fpeaks_["dt"]
                    recalc_ctime[~(fpeaks_["area"] > 0)] = 0.0
                # setting central times in the same way as inside peak processing
                mask = fpeaks_["area"] > 0
                result[f"{type_}_center_time_{arr_}"] = events[f"{type_}_time"]
                result[f"{type_}_center_time_{arr_}"][mask] += recalc_ctime[mask].astype(int)
                # computing widths ##
                # zero or undefined area peaks should have nans
                _, width, area_decile_from_midpoint = strax.compute_widths(fpeaks_)
                result[f"{type_}_rise_time_{arr_}"][:] = np.nan
                result[f"{type_}_rise_time_{arr_}"][mask] = -area_decile_from_midpoint[mask][:, 1]
                result[f"{type_}_range_50p_area_{arr_}"][:] = np.nan
                result[f"{type_}_range_50p_area_{arr_}"][mask] = width[mask][:, 5]
                result[f"{type_}_range_90p_area_{arr_}"][:] = np.nan
                result[f"{type_}_range_90p_area_{arr_}"][mask] = width[mask][:, 9]
            # Difference between center times of top and bottom arrays
            result[f"{type_}_center_time_diff_top_bot"] = (
                result[f"{type_}_center_time_top"] - result[f"{type_}_center_time_bot"]
            )
        return result



import strax
from ._peak_positions_base import PeakPositionsBase

export, __all__ = strax.exporter()


@export
class PeakPositionsCNF(PeakPositionsBase):

    __version__ = "0.0.0"
    child_plugin = True
    algorithm = "cnf"
    depends_on = (
        "peaklet_positions_cnf",
        "enhanced_peaklet_classification",
        "merged_s2s",
        "merged_s2_positions_cnf",
    )
    provides = "peak_positions_cnf"



from enum import IntEnum
import numpy as np
import strax
import straxen

from straxen.plugins.raw_records.daqreader import ARTIFICIAL_DEADTIME_CHANNEL

export, __all__ = strax.exporter()


# More info about the acquisition monitor can be found here:
# https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenon1t:alexelykov:acquisition_monitor


@export
class AqmonChannels(IntEnum):
    """Mapper of named aqmon channels to ints."""

    MV_TRIGGER = 797
    GPS_SYNC = 798
    ARTIFICIAL_DEADTIME = ARTIFICIAL_DEADTIME_CHANNEL
    # Analogue sum waveform
    SUM_WF = 800
    # GPS sync acquisition monitor
    GPS_SYNC_AM = 801
    # HighEnergyVeto
    HEV_STOP = 802
    HEV_START = 803
    # To avoid confusion with HEV, these are the high energy boards
    BUSY_HE_STOP = 804
    BUSY_HE_START = 805
    # Low energy boards (main chain)
    BUSY_STOP = 806
    BUSY_START = 807
    # nVETO GPS_Sync
    GPS_SYNC_NV = 813
    # mVETO GPS_Sync
    GPS_SYNC_MV = 1084


@export
class AqmonHits(strax.Plugin):
    """Find hits in acquisition monitor data.

    These hits could be then used by other plugins for deadtime calculations, GPS SYNC analysis,
    etc.

    """

    save_when = strax.SaveWhen.TARGET
    __version__ = "1.1.2"
    hit_min_amplitude_aqmon = straxen.URLConfig(
        default=(
            # Analogue signals
            (50, (int(AqmonChannels.SUM_WF),)),
            # Digital signals, can set a much higher threshold
            (
                1500,
                (
                    int(AqmonChannels.MV_TRIGGER),
                    int(AqmonChannels.GPS_SYNC),
                    int(AqmonChannels.GPS_SYNC_AM),
                    int(AqmonChannels.HEV_STOP),
                    int(AqmonChannels.HEV_START),
                    int(AqmonChannels.BUSY_HE_STOP),
                    int(AqmonChannels.BUSY_HE_START),
                    int(AqmonChannels.BUSY_STOP),
                    int(AqmonChannels.BUSY_START),
                ),
            ),
            # Fake signals, 0 meaning that we won't find hits using
            # strax but just look for starts and stops
            (0, (int(AqmonChannels.ARTIFICIAL_DEADTIME),)),
        ),
        track=True,
        help=(
            "Minimum hit threshold in ADC*counts above baseline. Specified "
            "per channel in the format (threshold, (chx,chy),)"
        ),
    )
    baseline_samples_aqmon = straxen.URLConfig(
        default=10,
        track=True,
        help="Number of samples to use at the start of the pulse to determine the baseline",
    )
    check_raw_record_aqmon_overlaps = straxen.URLConfig(
        default=True,
        track=False,
        help=(
            "Crash if any of the pulses in raw_records_aqmon overlap with others "
            "in the same channel"
        ),
    )

    depends_on = "raw_records_aqmon"
    provides = "aqmon_hits"
    data_kind = "aqmon_hits"

    dtype = strax.hit_dtype

    def compute(self, raw_records_aqmon):
        not_allowed_channels = set(np.unique(raw_records_aqmon["channel"])) - set(
            self.aqmon_channels
        )
        if not_allowed_channels:
            raise ValueError(
                f"Unknown channel {not_allowed_channels}. Only know {self.aqmon_channels}"
            )

        if self.check_raw_record_aqmon_overlaps:
            straxen.check_overlaps(raw_records_aqmon, n_channels=max(AqmonChannels).value + 1)

        records = strax.raw_to_records(raw_records_aqmon)
        strax.zero_out_of_bounds(records)
        strax.baseline(records, baseline_samples=self.baseline_samples_aqmon, flip=True)
        aqmon_hits = self.find_aqmon_hits_per_channel(records)
        aqmon_hits = strax.sort_by_time(aqmon_hits)
        return aqmon_hits

    @property
    def aqmon_channels(self):
        return [
            channel
            for hit_and_channel_list in self.hit_min_amplitude_aqmon
            for channel in hit_and_channel_list[1]
        ]

    def find_aqmon_hits_per_channel(self, records):
        """Allow different thresholds to be applied to different channels."""
        aqmon_thresholds = np.zeros(np.max(self.aqmon_channels) + 1)
        for hit_threshold, channels in self.hit_min_amplitude_aqmon:
            aqmon_thresholds[np.array(channels)] = hit_threshold

        # Split the artificial deadtime ones and do those separately if there are any
        is_artificial = records["channel"] == AqmonChannels.ARTIFICIAL_DEADTIME
        aqmon_hits = strax.find_hits(records[~is_artificial], min_amplitude=aqmon_thresholds)

        if np.sum(is_artificial):
            aqmon_hits = np.concatenate(
                [aqmon_hits, self.get_deadtime_hits(records[is_artificial])]
            )
        return aqmon_hits

    def get_deadtime_hits(self, artificial_deadtime):
        """Actually, the artificial deadtime hits are already an interval so we only have to copy
        the appropriate hits."""
        hits = np.zeros(len(artificial_deadtime), dtype=self.dtype)
        hits["time"] = artificial_deadtime["time"]
        hits["dt"] = artificial_deadtime["dt"]
        hits["length"] = artificial_deadtime["length"]
        hits["channel"] = artificial_deadtime["channel"]
        return hits
